import traceback

import sqlalchemy as sa

from typing import List, Dict, Text, Type, Set, Tuple
from threading import RLock

from geoalchemy2 import Geometry as GeoalchemyGeometry
from sqlalchemy.sql import bindparam
from appyratus.env import Environment

from ravel.query.predicate import (
    Predicate, ConditionalPredicate, BooleanPredicate,
    OP_CODE,
)
from ravel.schema import fields, Field
from ravel.util.loggers import console
from ravel.util.json_encoder import JsonEncoder
from ravel.util import get_class_name
from ravel.store.base import Store
from ravel.constants import REV, ID

from .dialect import Dialect
from .sqlalchemy_table_builder import SqlalchemyTableBuilder
from ..types import ArrayOfEnum, UtcDateTime
from ..postgis import (
    POSTGIS_OP_CODE,
    GeometryObject, PointGeometry, PolygonGeometry,
    Point, Polygon,
)

json_encoder = JsonEncoder()


class SqlalchemyStore(Store):
    """
    A SQLAlchemy-based store, which keeps a single connection pool (AKA
    Engine) shared by all threads; however, each thread keeps singleton
    thread-local database connection and transaction objects, managed through
    connect()/close() and begin()/end().
    """

    env = Environment(
        SQLALCHEMY_STORE_ECHO=fields.Bool(default=False),
        SQLALCHEMY_STORE_SHOW_QUERIES=fields.Bool(default=False),
        SQLALCHEMY_STORE_DIALECT=fields.Enum(
            fields.String(), Dialect.values(), default=Dialect.sqlite
        ),
        SQLALCHEMY_STORE_PROTOCOL=fields.String(default='sqlite'),
        SQLALCHEMY_STORE_USER=fields.String(),
        SQLALCHEMY_STORE_HOST=fields.String(),
        SQLALCHEMY_STORE_PORT=fields.String(),
        SQLALCHEMY_STORE_NAME=fields.String(),
    )

    # id_column_names is a mapping from table name to its _id column name
    _id_column_names = {}

    # only one thread needs to bootstrap the SqlalchemyStore. This lock is
    # used to ensure that this is what happens when the host app bootstraps.
    _bootstrap_lock = RLock()

    @classmethod
    def get_default_adapters(cls, dialect: Dialect, table_name) -> List[Field.Adapter]:
        # TODO: Move this into the adapters file

        adapters = [
            fields.Field.adapt(
                on_adapt=lambda field: sa.Text,
                on_encode=lambda x: cls.ravel.app.json.encode(x),
                on_decode=lambda x: cls.ravel.app.json.decode(x),
            ),
            fields.Email.adapt(on_adapt=lambda field: sa.Text),
            fields.Bytes.adapt(on_adapt=lambda field: sa.LargeBinary),
            fields.BcryptString.adapt(on_adapt=lambda field: sa.Text),
            fields.Float.adapt(on_adapt=lambda field: sa.Float),
            fields.DateTime.adapt(on_adapt=lambda field: UtcDateTime),
            fields.Timestamp.adapt(on_adapt=lambda field: UtcDateTime),
            fields.Bool.adapt(on_adapt=lambda field: sa.Boolean),
            fields.TimeDelta.adapt(on_adapt=lambda field: sa.Interval),
            fields.Enum.adapt(
                on_adapt=lambda field: {
                    fields.String: sa.Text,
                    fields.Int: sa.Integer,
                    fields.Float: sa.Float,
                }[type(field.nested)]
            ),
        ]
        adapters.extend(
            field_class.adapt(on_adapt=lambda field: sa.Text)
            for field_class in {
                fields.String, fields.FormatString,
                fields.UuidString, fields.DateTimeString
            }
        )
        adapters.extend(
            field_class.adapt(on_adapt=lambda field: sa.BigInteger)
            for field_class in {
                fields.Int, fields.Uint32, fields.Uint64, fields.Uint,
                fields.Int32,
            }
        )
        if dialect == Dialect.postgresql:
            adapters.extend(cls.get_postgresql_default_adapters(table_name))
        elif dialect == Dialect.mysql:
            adapters.extend(cls.get_mysql_default_adapters(table_name))
        elif dialect == Dialect.sqlite:
            adapters.extend(cls.get_sqlite_default_adapters(table_name))

        return adapters

    @classmethod
    def get_postgresql_default_adapters(cls, table_name) -> List[Field.Adapter]:
        pg_types = sa.dialects.postgresql

        def on_adapt_list(field):
            if isinstance(field.nested, fields.Enum):
                name = f'{table_name}__{field.name}'
                return ArrayOfEnum(
                    pg_types.ENUM(*field.nested.values, name=name)
                )
            return pg_types.ARRAY({
                fields.String: sa.Text,
                fields.Email: sa.Text,
                fields.Uuid: pg_types.UUID,
                fields.Int: sa.Integer,
                fields.Bool: sa.Boolean,
                fields.Float: sa.Float,
                fields.DateTime: UtcDateTime,
                fields.Timestamp: UtcDateTime,
                fields.Dict: pg_types.JSONB,
                fields.Field: pg_types.JSONB,
                fields.Nested: pg_types.JSONB,
            }.get(type(field.nested), sa.Text))

        return [
            Point.adapt(
                on_adapt=lambda field: GeoalchemyGeometry(field.geo_type),
                on_encode=lambda x: x.to_EWKT_string(),
                on_decode=lambda x: (
                    PointGeometry(x['geometry']['coordinates']) if x
                    else None
                )
            ),
            Polygon.adapt(
                on_adapt=lambda field: GeoalchemyGeometry(field.geo_type),
                on_encode=lambda x: x.to_EWKT_string(),
                on_decode=lambda x: PolygonGeometry(
                    x['geometry']['coordinates'] if x
                    else None
                )
            ),
            fields.Field.adapt(on_adapt=lambda field: pg_types.JSONB),
            fields.Uuid.adapt(on_adapt=lambda field: pg_types.UUID),
            fields.Dict.adapt(on_adapt=lambda field: pg_types.JSONB),
            fields.Nested.adapt(
                on_adapt=lambda field: pg_types.JSONB,
            ),
            fields.Set.adapt(
                on_adapt=lambda field: pg_types.JSONB,
                on_encode=lambda x: list(x),
                on_decode=lambda x: set(x)
            ),
            fields.UuidString.adapt(
                on_adapt=lambda field: pg_types.UUID,
                on_decode=lambda x: x.replace('-', '') if x else x,
            ),
            fields.List.adapt(on_adapt=on_adapt_list)
        ]

    @classmethod
    def get_mysql_default_adapters(cls, table_name) -> List[Field.Adapter]:
        return [
            fields.Dict.adapt(on_adapt=lambda field: sa.JSON),
            fields.Nested.adapt(on_adapt=lambda field: sa.JSON),
            fields.List.adapt(on_adapt=lambda field: sa.JSON),
            fields.Set.adapt(
                on_adapt=lambda field: sa.JSON,
                on_encode=lambda x: cls.ravel.app.json.encode(x),
                on_decode=lambda x: set(cls.ravel.app.json.decode(x))
            ),
        ]

    @classmethod
    def get_sqlite_default_adapters(cls, table_name) -> List[Field.Adapter]:
        adapters = [
            field_class.adapt(
                on_adapt=lambda field: sa.Text,
                on_encode=lambda x: cls.ravel.app.json.encode(x),
                on_decode=lambda x: cls.ravel.app.json.decode(x),
            )
            for field_class in {
                fields.Dict, fields.List, fields.Nested
            }
        ]
        adapters.append(
            fields.Set.adapt(
                on_adapt=lambda field: sa.Text,
                on_encode=lambda x: cls.ravel.app.json.encode(x),
                on_decode=lambda x: set(cls.ravel.app.json.decode(x))
            )
        )
        return adapters

    def __init__(self, adapters: List[Field.Adapter] = None):
        super().__init__()
        self._custom_adapters = adapters or []
        self._table = None
        self._builder = None
        self._adapters = None
        self._id_column = None
        self._options = {}

    @property
    def adapters(self):
        return self._adapters

    @property
    def id_column_name(self):
        return self.resource_type.Schema.fields[ID].source

    def prepare(self, record: Dict, serialize=True) -> Dict:
        """
        When inserting or updating data, the some raw values in the record
        dict must be transformed before their corresponding sqlalchemy column
        type will accept the data.
        """
        cb_name = 'on_encode' if serialize else 'on_decode'
        prepared_record = {}
        for k, v in record.items():
            if k in (REV):
                prepared_record[k] = v
            adapter = self._adapters.get(k)
            if adapter:
                callback = getattr(adapter, cb_name, None)
                if callback:
                    try:
                        prepared_record[k] = callback(v)
                        continue
                    except Exception:
                        console.error(
                            message=f'failed to adapt column value: {k}',
                            data={
                                'value': v,
                                'field': adapter.field_class
                            }
                        )
                        raise
            prepared_record[k] = v
        return prepared_record

    def adapt_id(self, _id, serialize=True):
        cb_name = 'on_encode' if serialize else 'on_decode'
        adapter = self._adapters.get(self.id_column_name)
        if adapter:
            callback = getattr(adapter, cb_name)
            if callback:
                return callback(_id)
        return _id

    @classmethod
    def on_bootstrap(cls, url=None, dialect=None, echo=False, db=None, **kwargs):
        """
        Initialize the SQLAlchemy connection pool (AKA Engine).
        """
        with cls._bootstrap_lock:
            cls.ravel.kwargs = kwargs

            # construct the URL to the DB server
            # url can be a string or a dict
            if isinstance(url, dict):
                url_parts = url
                cls.ravel.app.shared.sqla_url = (
                    '{protocol}://{user}@{host}:{port}/{db}'.format(
                        **url_parts
                    )
                )
            elif isinstance(url, str):
                cls.ravel.app.shared.sqla_url = url
            else:
                url_parts = dict(
                    protocol=cls.env.SQLALCHEMY_STORE_PROTOCOL,
                    user=cls.env.SQLALCHEMY_STORE_USER or '',
                    host=(
                        '@' + cls.env.SQLALCHEMY_STORE_HOST
                        if cls.env.SQLALCHEMY_STORE_HOST else ''
                    ),
                    port=(
                        ':' + cls.env.SQLALCHEMY_STORE_PORT
                        if cls.env.SQLALCHEMY_STORE_PORT else ''
                    ),
                    db=(
                        '/' + (db or cls.env.SQLALCHEMY_STORE_NAME or '')
                    )
                )
                cls.ravel.app.shared.sqla_url = url or (
                    '{protocol}://{user}{host}{port}{db}'.format(
                        **url_parts
                    )
                )

            cls.dialect = dialect or cls.env.SQLALCHEMY_STORE_DIALECT

            from sqlalchemy.dialects import postgresql, sqlite, mysql

            cls.sa_dialect = None
            if cls.dialect == Dialect.postgresql:
                cls.sa_dialect = postgresql
            elif cls.dialect == Dialect.sqlite:
                cls.sa_dialect = sqlite
            elif cls.dialect == Dialect.mysql:
                cls.sa_dialect = mysql

            console.debug(
                message='creating sqlalchemy engine',
                data={
                    'echo': cls.env.SQLALCHEMY_STORE_ECHO,
                    'dialect': cls.dialect,
                    'url': cls.ravel.app.shared.sqla_url,
                }
            )

            cls.ravel.local.sqla_tx = None
            cls.ravel.local.sqla_conn = None
            cls.ravel.local.sqla_metadata = sa.MetaData()
            cls.ravel.local.sqla_metadata.bind = sa.create_engine(
                name_or_url=cls.ravel.app.shared.sqla_url,
                echo=bool(echo or cls.env.SQLALCHEMY_STORE_ECHO),
                strategy='threadlocal'
            )

            # set global thread-local sqlalchemy store method aliases
            cls.ravel.app.local.create_tables = cls.create_tables

    def on_bind(
        self,
        resource_type: Type['Resource'],
        table: Text = None,
        schema: 'Schema' = None,
        **kwargs
    ):
        """
        Initialize SQLAlchemy data strutures used for constructing SQL
        expressions used to manage the bound resource type.
        """
        # map each of the resource's schema fields to a corresponding adapter,
        # which is used to prepare values upon insert and update.
        table = (
            table or SqlalchemyTableBuilder.derive_table_name(resource_type)
        )
        field_class_2_adapter = {
            adapter.field_class: adapter for adapter in
            self.get_default_adapters(self.dialect, table) + self._custom_adapters
        }
        self._adapters = {
            field_name: field_class_2_adapter[type(field)]
            for field_name, field in self.resource_type.Schema.fields.items()
            if (
                type(field) in field_class_2_adapter and
                field.meta.get('ravel_on_resolve') is None
            )
        }

        # build the Sqlalchemy Table object for the bound resource type.
        self._builder = SqlalchemyTableBuilder(self)

        try:
            self._table = self._builder.build_table(name=table, schema=schema)
        except Exception:
            console.error(f'failed to build sa.Table: {table}')
            raise

        self._id_column = getattr(self._table.c, self.id_column_name)

        # remember which column is the _id column
        self._id_column_names[self._table.name] = self.id_column_name

        # set SqlalchemyStore options here, using bootstrap-level
        # options as base/default options.
        self._options = dict(self.ravel.kwargs, **kwargs)

    def query(
        self,
        predicate: 'Predicate',
        fields: Set[Text] = None,
        limit: int = None,
        offset: int = None,
        order_by: Tuple = None,
        **kwargs,
    ):
        fields = fields or {
            k: None for k in self._adapters
        }
        fields.update({
            self.id_column_name: None,
            self.resource_type.Schema.fields[REV].source: None,
        })

        columns = []
        table_alias = self.table.alias(
            ''.join(s.strip('_')[0] for s in self.table.name.split('_'))
        )
        for k in fields:
            col = getattr(table_alias.c, k)
            if isinstance(col.type, GeoalchemyGeometry):
                columns.append(sa.func.ST_AsGeoJSON(col).label(k))
            else:
                columns.append(col)

        predicate = Predicate.deserialize(predicate)
        filters = self._prepare_predicate(table_alias, predicate)

        # build the query object
        query = sa.select(columns).where(filters)

        if order_by:
            sa_order_by = [
                sa.desc(getattr(table_alias.c, x.key)) if x.desc else
                sa.asc(getattr(table_alias.c, x.key))
                for x in order_by
            ]
            query = query.order_by(*sa_order_by)

        if limit is not None:
            query = query.limit(max(0, limit))
        if offset is not None:
            query = query.offset(max(0, limit))

        console.debug(
            message=(
                f'SQL: SELECT FROM {self.table}'
                + (f' OFFSET {offset}' if offset is not None else '')
                + (f' LIMIT {limit}' if limit else '')
                + (f' ORDER BY {", ".join(x.to_sql() for x in order_by)}'
                    if order_by else '')
            ),
            data={
                'stack': traceback.format_stack(),
                'statement': str(query.compile(
                    compile_kwargs={'literal_binds': True}
                )).split('\n')
            }
            if self.env.SQLALCHEMY_STORE_SHOW_QUERIES
            else None
        )
        # execute query, aggregating resulting records
        cursor = self.conn.execute(query)
        records = []

        while True:
            page = [
                self.prepare(dict(row.items()), serialize=False)
                for row in cursor.fetchmany(512)
            ]
            if page:
                records.extend(page)
            else:
                break

        return records

    def _prepare_predicate(self, table, pred, empty=set()):
        if isinstance(pred, ConditionalPredicate):
            if not pred.ignore_field_adapter:
                adapter = self._adapters.get(pred.field.source)
                if adapter and adapter.on_encode:
                    pred.value = adapter.on_encode(pred.value)
            col = getattr(table.c, pred.field.source)
            if pred.op == OP_CODE.EQ:
                return col == pred.value
            elif pred.op == OP_CODE.NEQ:
                return col != pred.value
            if pred.op == OP_CODE.GEQ:
                return col >= pred.value
            elif pred.op == OP_CODE.GT:
                return col > pred.value
            elif pred.op == OP_CODE.LT:
                return col < pred.value
            elif pred.op == OP_CODE.LEQ:
                return col <= pred.value
            elif pred.op == OP_CODE.INCLUDING:
                return col.in_(pred.value)
            elif pred.op == OP_CODE.EXCLUDING:
                return ~col.in_(pred.value)
            elif pred.op == POSTGIS_OP_CODE.CONTAINS:
                if isinstance(pred.value, GeometryObject):
                    EWKT_str = pred.value.to_EWKT_string()
                else:
                    EWKT_str = pred.value
                return sa.func.ST_Contains(
                    col, sa.func.ST_GeomFromEWKT(EWKT_str),
                )
            elif pred.op == POSTGIS_OP_CODE.CONTAINED_BY:
                if isinstance(pred.value, GeometryObject):
                    EWKT_str = pred.value.to_EWKT_string()
                else:
                    EWKT_str = pred.value
                return sa.func.ST_Contains(
                    sa.func.ST_GeomFromEWKT(EWKT_str), col
                )
            elif pred.op == POSTGIS_OP_CODE.WITHIN_RADIUS:
                center = pred.value['center']
                radius = pred.value['radius']
                return sa.func.ST_PointInsideCircle(
                    col, center[0], center[1], radius
                )
            else:
                raise Exception('unrecognized conditional predicate')
        elif isinstance(pred, BooleanPredicate):
            if pred.op == OP_CODE.AND:
                lhs_result = self._prepare_predicate(table, pred.lhs)
                rhs_result = self._prepare_predicate(table, pred.rhs)
                return sa.and_(lhs_result, rhs_result)
            elif pred.op == OP_CODE.OR:
                lhs_result = self._prepare_predicate(table, pred.lhs)
                rhs_result = self._prepare_predicate(table, pred.rhs)
                return sa.or_(lhs_result, rhs_result)
            else:
                raise Exception('unrecognized boolean predicate')
        else:
            raise Exception('unrecognized predicate type')

    def exists(self, _id) -> bool:
        columns = [sa.func.count(self._id_column)]
        query = (
            sa.select(columns).where(
                self._id_column == self.adapt_id(_id)
            )
        )
        result = self.conn.execute(query)
        return bool(result.scalar())

    def exists_many(self, _ids: Set) -> Dict[object, bool]:
        columns = [self._id_column, sa.func.count(self._id_column)]
        query = (
            sa.select(columns).where(
                self._id_column.in_(
                    [self.adapt_id(_id) for _id in _ids]
                )
            )
        )
        return {
            row[0]: row[1] for row in self.conn.execute(query)
        }

    def count(self) -> int:
        query = sa.select([sa.func.count(self._id_column)])
        result = self.conn.execute(query)
        return result.scalar()

    def fetch(self, _id, fields=None) -> Dict:
        records = self.fetch_many(_ids=[_id], fields=fields)
        return records[_id] if records else None

    def fetch_many(self, _ids: List, fields=None, as_list=False) -> Dict:
        prepared_ids = [self.adapt_id(_id, serialize=True) for _id in _ids]

        if fields:
            if not isinstance(fields, set):
                fields = set(fields)
        else:
            fields = {
                f.source for f in self.resource_type.Schema.fields.values()
                if f.name in self._adapters
            }
        fields.update({
            self.id_column_name,
            self.resource_type.Schema.fields[REV].source,
        })

        columns = []
        for k in fields:
            col = getattr(self.table.c, k)
            if isinstance(col.type, GeoalchemyGeometry):
                columns.append(sa.func.ST_AsGeoJSON(col).label(k))
            else:
                columns.append(col)

        select_stmt = sa.select(columns)

        id_col = getattr(self.table.c, self.id_column_name)

        if prepared_ids:
            select_stmt = select_stmt.where(id_col.in_(prepared_ids))
        cursor = self.conn.execute(select_stmt)
        records = {} if not as_list else []

        while True:
            page = cursor.fetchmany(512)
            if page:
                for row in page:
                    raw_record = dict(row.items())
                    record = self.prepare(raw_record, serialize=False)
                    _id = self.adapt_id(
                        row[self.id_column_name], serialize=False
                    )
                    if as_list:
                        records.append(record)
                    else:
                        records[_id] = record
            else:
                break

        return records

    def fetch_all(self, fields: Set[Text] = None) -> Dict:
        return self.fetch_many([], fields=fields)

    def create(self, record: dict) -> dict:
        record[self.id_column_name] = self.create_id(record)
        prepared_record = self.prepare(record, serialize=True)
        insert_stmt = self.table.insert().values(**prepared_record)
        _id = prepared_record.get('_id', '')
        console.debug(
            f'SQL: INSERT {str(_id)[:7] + " " if _id else ""}'
            f'INTO {self.table}'
        )
        try:
            if self.supports_returning:
                insert_stmt = insert_stmt.return_defaults()
                result = self.conn.execute(insert_stmt)
                return dict(record, **(result.returned_defaults or {}))
            else:
                result = self.conn.execute(insert_stmt)
                return self.fetch(_id=record[self.id_column_name])
        except Exception:
            console.error(
                message=f'failed to insert record',
                data={
                    'record': record,
                    'resource': get_class_name(self.resource_type),
                }
            )
            raise

    def create_many(self, records: List[Dict]) -> Dict:
        prepared_records = []
        nullable_fields = self.resource_type.Schema.nullable_fields
        for record in records:
            record[self.id_column_name] = self.create_id(record)
            prepared_record = self.prepare(record, serialize=True)
            prepared_records.append(prepared_record)
            for nullable_field in nullable_fields.values():
                if nullable_field.name not in prepared_record:
                    prepared_record[nullable_field.name] = None


        try:
            self.conn.execute(self.table.insert(), prepared_records)
        except Exception:
            console.error(f'failed to insert records')
            raise

        n = len(prepared_records)
        id_list_str = (
            ', '.join(str(x['_id'])[:7]
            for x in prepared_records if x.get('_id'))
        )
        console.debug(
            f'SQL: INSERT {id_list_str} INTO {self.table} '
            + (f'(count: {n})' if n > 1 else '')
        )

        if self.supports_returning:
            # TODO: use implicit returning if possible
            pass

        return self.fetch_many(
            (rec[self.id_column_name] for rec in records), as_list=True)

    def update(self, _id, data: Dict) -> Dict:
        prepared_id = self.adapt_id(_id)
        prepared_data = self.prepare(data, serialize=True)
        if prepared_data:
            update_stmt = (
                self.table
                    .update()
                    .values(**prepared_data)
                    .where(self._id_column == prepared_id)
                )
        else:
            return prepared_data
        if self.supports_returning:
            update_stmt = update_stmt.return_defaults()
            console.debug(f'SQL: UPDATE {self.table}')
            result = self.conn.execute(update_stmt)
            return dict(data, **(result.returned_defaults or {}))
        else:
            self.conn.execute(update_stmt)
            if self._options.get('fetch_on_update', True):
                return self.fetch(_id)
            return data

    def update_many(self, _ids: List, data: List[Dict] = None) -> None:
        assert data

        prepared_ids = []
        prepared_records = []

        for _id, record in zip(_ids, data):
            prepared_id = self.adapt_id(_id)
            prepared_record = self.prepare(record, serialize=True)
            if prepared_record:
                prepared_ids.append(prepared_id)
                prepared_records.append(prepared_record)
                prepared_record[ID] = prepared_id

        if prepared_records:
            n = len(prepared_records)
            console.debug(
                f'SQL: UPDATE {self.table} '
                + (f'({n}x)' if n > 1else '')
            )
            values = {
                k: bindparam(k) for k in prepared_records[0].keys()
            }
            update_stmt = (
                self.table
                    .update()
                    .where(self._id_column == bindparam(self.id_column_name))
                    .values(**values)
            )
            self.conn.execute(update_stmt, prepared_records)

        if self._options.get('fetch_on_update', True):
            if self.supports_returning:
                # TODO: use implicit returning if possible
                return self.fetch_many(_ids)
            else:
                return self.fetch_many(_ids)
        return

    def delete(self, _id) -> None:
        prepared_id = self.adapt_id(_id)
        delete_stmt = self.table.delete().where(
            self._id_column == prepared_id
        )
        self.conn.execute(delete_stmt)

    def delete_many(self, _ids: list) -> None:
        prepared_ids = [self.adapt_id(_id) for _id in _ids]
        delete_stmt = self.table.delete().where(
            self._id_column.in_(prepared_ids)
        )
        self.conn.execute(delete_stmt)

    def delete_all(self):
        delete_stmt = self.table.delete()
        self.conn.execute(delete_stmt)

    @property
    def table(self):
        return self._table

    @property
    def conn(self):
        sqla_conn = getattr(self.ravel.local, 'sqla_conn', None)
        if sqla_conn is None:
            # lazily initialize a connection for this thread
            self.connect()
        return self.ravel.local.sqla_conn

    @property
    def supports_returning(self):
        if not self.is_bootstrapped():
            return False
        metadata = self.get_metadata()
        return metadata.bind.dialect.implicit_returning

    @classmethod
    def create_tables(cls, overwrite=False):
        """
        Create all tables for all SqlalchemyStores used in the host app.
        """
        if not cls.is_bootstrapped():
            console.error(
                f'{get_class_name(cls)} cannot create '
                f'tables unless bootstrapped'
            )
            return

        meta = cls.get_metadata()
        engine = cls.get_engine()

        if overwrite:
            console.info('dropping Resource SQL tables...')
            meta.drop_all(engine)

        # create all tables
        console.info('creating Resource SQL tables...')
        meta.create_all(engine)

    @classmethod
    def get_active_connection(cls):
        return getattr(cls.ravel.local, 'sqla_conn', None)

    @classmethod
    def connect(cls, refresh=True):
        """
        Create a singleton thread-local SQLAlchemy connection, shared across
        all Resources backed by a SQLAlchemy store. When working with multiple
        threads or processes, make sure to 
        """
        sqla_conn = getattr(cls.ravel.local, 'sqla_conn', None)
        metadata = cls.ravel.local.sqla_metadata
        if sqla_conn is not None:
            console.warning(
                message='sqlalchemy store already has connection',
            )
            if refresh:
                cls.close()
                cls.ravel.local.sqla_conn = metadata.bind.connect()
        else:
            cls.ravel.local.sqla_conn = metadata.bind.connect()

        return cls.ravel.local.sqla_conn

    @classmethod
    def close(cls):
        """
        Return the thread-local database connection to the sqlalchemy
        connection pool (AKA the "engine").
        """
        sqla_conn = getattr(cls.ravel.local, 'sqla_conn', None)
        if sqla_conn is not None:
            console.debug('closing sqlalchemy connection')
            sqla_conn.close()
            cls.ravel.local.sqla_conn = None
        else:
            console.warning('sqlalchemy has no connection to close')

    @classmethod
    def begin(cls, auto_connect=True, **kwargs):
        """
        Initialize a thread-local transaction. An exception is raised if
        there's already a pending transaction.
        """
        conn = cls.get_active_connection()
        if conn is None:
            if auto_connect:
                conn = cls.connect()
            else:
                raise Exception('no active sqlalchemy connection')

        existing_tx = getattr(cls.ravel.local, 'sqla_tx', None)
        if existing_tx is not None:
            console.debug('there is already an open transaction')
        else:
            new_tx = cls.ravel.local.sqla_conn.begin()
            cls.ravel.local.sqla_tx = new_tx

    @classmethod
    def commit(cls, rollback=True, **kwargs):
        """
        Call commit on the thread-local database transaction. "Begin" must be
        called to start a new transaction at this point, if a new transaction
        is desired.
        """
        def perform_sqlalchemy_commit():
            tx = getattr(cls.ravel.local, 'sqla_tx', None)
            if tx is not None:
                cls.ravel.local.sqla_tx.commit()
                cls.ravel.local.sqla_tx = None

        # try to commit the transaction.
        console.debug(f'committing sqlalchemy transaction')
        try:
            perform_sqlalchemy_commit()
        except Exception:
            if rollback:
                # if the commit fails, rollback the transaction
                console.critical(
                    f'rolling back sqlalchemy transaction'
                )
                cls.rollback()
            else:
                console.exception(
                    f'sqlalchemy transaction failed commit'
                )
        finally:
            # ensure we close the connection either way
            cls.close()

    @classmethod
    def rollback(cls, **kwargs):
        tx = getattr(cls.ravel.local, 'sqla_tx', None)
        if tx is not None:
            cls.ravel.local.sqla_tx = None
            try:
                tx.rollback()
            except:
                console.exception(
                    f'sqlalchemy transaction failed to rollback'
                )

    @classmethod
    def has_transaction(cls) -> bool:
        return cls.ravel.local.sqla_tx is not None

    @classmethod
    def get_metadata(cls):
        return cls.ravel.local.sqla_metadata

    @classmethod
    def get_engine(cls):
        return cls.get_metadata().bind

    @classmethod
    def dispose(cls):
        meta = cls.get_metadata()
        if not meta:
            cls.ravel.local.sqla_metadata = None
            return

        engine = meta.bind
        engine.dispose()



from __future__ import absolute_import

from ravel.util.json_encoder import JsonEncoder

from falcon import HTTPError, status_codes

json = JsonEncoder()


class RavelFalconError(HTTPError):

    http_status = status_codes.HTTP_400
    api_status = 'bad-request'
    default_message = 'bad request'

    def __init__(self, message=None, data=None):
        self.data = data or {}
        self.message = message or self.default_message
        super().__init__(status=self.http_status, description=self.message)

    def __repr__(self):
        return '<FalconError({})>'.format(self.__class__.__name__)

    def to_dict(self):
        return {
            'status': self.http_status,
            'message': self.message,
            'data': self.data,
            }

    def to_json(self):
        return json.encode(self.to_dict())


class NotFound(RavelFalconError):

    http_status = status_codes.HTTP_404
    api_status = 'not-found'
    default_message = 'resource not found'

    def __init__(self, resource_name, data=None):
        super().__init__(
            message='{} not found'.format(resource_name),
            data=data
            )


class NotAuthenticated(RavelFalconError):

    http_status = status_codes.HTTP_401
    api_status = 'not-authenticated'
    default_message = 'unauthenticated'

    def __init__(self):
        super().__init__()



from typing import Text, Tuple, List, Set, Dict, Type, Union, Callable
from random import randint

from ravel.util.loggers import console
from ravel.util.misc_functions import get_class_name, flatten_sequence
from ravel.query.predicate import (
    Predicate,
    ConditionalPredicate,
    BooleanPredicate,
    OP_CODE,
)


from ravel.constants import ID
from ravel.util import is_resource, is_batch
from ravel.query.order_by import OrderBy
from ravel.query.request import Request
from ravel.resolver.resolver import Resolver
from ravel.resolver.resolver_property import ResolverProperty
from ravel.schema import fields


class Loader(Resolver):
    """
    The Loader resolver is responsible for fetching Resource fields.
    """

    def __init__(self, field=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._field = None
        if field:
            self.field = field

    @classmethod
    def property_type(cls):
        return LoaderProperty

    @classmethod
    def tags(cls) -> Set[Text]:
        return {'fields'}

    @classmethod
    def priority(cls) -> int:
        return 1

    @property
    def field(self) -> 'Field':
        return self._field

    @property
    def is_virtual(self) -> bool:
        return bool(self.field.meta.get('ravel_on_resolve'))

    @field.setter
    def field(self, field: 'Field'):
        self._field = field
        self.nullable = field.nullable
        self.required = field.required
        self.private = field.meta.get('private', False)

    def on_copy(self, copy):
        copy.field = self.field

    def on_resolve(self, resource, request):
        exists_resource = ID in resource.internal.state
        if not exists_resource:
            return None

        # get all non-virtual field names in resource schema
        all_field_names = {
            f.name for f in resource.ravel.schema.fields.values()
            if not f.meta.get('ravel_on_resolve')
        }

        # of those, take only thos which are not yet loaded from the store.
        # these are the ones we are going to try to fetch.
        unloaded_field_names = (
            all_field_names - resource.internal.state.keys()
        )
        unloaded_field_names.add(self._field.name)
        
        # container for resolved resource instance state
        new_resource_state = {}

        # if the field has a on_resolve function, put there by the
        # @field decorator, use it instead of fetching from store.
        field_on_resolve = self._field.meta.get('ravel_on_resolve')
        if field_on_resolve is not None:
            unloaded_field_names.discard(self._field.name)
            value = field_on_resolve(resource, request)
            new_resource_state[self._field.name] = value

        # merge new state into existing resoruce instance state
        new_resource_state.update(
            resource.ravel.local.store.dispatch(
                'fetch',
                args=(resource._id, ),
                kwargs={'fields': unloaded_field_names.copy()}
            ) or {}
        )

        # merge in new state to existing resource, not overwriting
        # any fields that are dirty, i.e. have changes.
        keys_to_clean = set()
        for k, v in new_resource_state.items():
            is_dirty = k in resource.internal.state.dirty
            if not is_dirty or k in unloaded_field_names:
                keys_to_clean.add(k)
                resource[k] = v

        # mark the loaded fields as "clean", meaning, we are telling the system
        # that thse fields are new, not stale and in need of saving.
        resource.clean(keys_to_clean)

        return resource

    def on_resolve_batch(self, batch, request):
        id_2_resource = {res._id: res for res in batch}
        resource_ids = []
        for res in batch:
            id_2_resource[res._id] = res
            res_id = res.internal.state.get('_id')
            if res_id is not None:
                resource_ids.append(res_id)

        # field names to fetch (fetch all eagerly)
        field_names = set(self.target.ravel.schema.fields.keys())
        state_dicts = self.owner.ravel.local.store.dispatch('fetch_many',
            args=(resource_ids, ),
            kwargs={'fields': field_names}
        )

        for res_id, state in state_dicts.items():
            if state:
                id_2_resource[res_id].merge(state)

        return batch

    def on_simulate(self, resource, request):
        value = None

        if self.nullable and self._field.name != ID:
            # use None as the simulated value 10% of the time
            if randint(1, 10) > 1:
                value = self._field.generate()
        else:
            value = self._field.generate()

        return value

    def on_backfill(self, resource, request):
        return self.on_simulate(resource, request)


class View(Loader):

    @classmethod
    def tags(cls) -> Set[Text]:
        return {'view'}

    @classmethod
    def priority(cls) -> int:
        return 20

    def on_resolve(self, resource, request):
        raise NotImplementedError()


class LoaderProperty(ResolverProperty):
    def __hash__(self):
        return super().__hash__()

    def __repr__(self):
        return f'{get_class_name(self.resolver.owner)}.{self.resolver.name}'

    def __eq__(self, other: Predicate) -> Predicate:
        return ConditionalPredicate(OP_CODE.EQ, self, other)

    def __ne__(self, other: Predicate) -> Predicate:
        return ConditionalPredicate(OP_CODE.NEQ, self, other)

    def __lt__(self, other: Predicate) -> Predicate:
        return ConditionalPredicate(OP_CODE.LT, self, other)

    def __le__(self, other: Predicate) -> Predicate:
        return ConditionalPredicate(OP_CODE.LEQ, self, other)

    def __gt__(self, other: Predicate) -> Predicate:
        return ConditionalPredicate(OP_CODE.GT, self, other)

    def __ge__(self, other: Predicate) -> Predicate:
        return ConditionalPredicate(OP_CODE.GEQ, self, other)

    def including(self, *others) -> Predicate:
        others = flatten_sequence(others)
        visited = set()
        deduplicated = []
        for obj in others:
            if is_resource(obj):
                if obj._id not in visited:
                    visited.add(obj._id)
                    deduplicated.append(obj._id)
            else:
                if obj not in visited:
                    visited.add(obj)
                    deduplicated.append(obj)
                
        return ConditionalPredicate(OP_CODE.INCLUDING, self, deduplicated)

    def excluding(self, *others) -> Predicate:
        others = flatten_sequence(others)
        others = {obj._id if is_resource(obj) else obj for obj in others}
        return ConditionalPredicate(OP_CODE.EXCLUDING, self, others)

    def fset(self, owner: 'Resource', value):
        field = self.resolver.field
        if value is None:
            super().fset(owner, None)
            return

        processed_value, errors = field.process(value)
        if errors:
            console.error(
                message=f'cannot set {self}',
                data={
                    'resolver': str(self.resolver),
                    'field': field,
                    'schema': owner.ravel.schema,
                    'errors': errors,
                    'value': value,
                }
            )
            raise Exception(str(errors))

        super().fset(owner, processed_value)

    @property
    def asc(self):
        return OrderBy(self.resolver.field.name, desc=False)

    @property
    def desc(self):
        return OrderBy(self.resolver.field.name, desc=True)



from base64 import b64decode, b64encode
from uuid import uuid4
from datetime import datetime, timedelta
from typing import Text, Tuple, Type, Dict

from Crypto.Cipher import AES
from Crypto.Random import get_random_bytes
from Crypto.Util.Padding import pad, unpad

from ravel.app.middleware import Middleware
from ravel.util.loggers import console
from ravel.ext.falcon.service import FalconService


class ManageCsrfToken(Middleware):
    """
    This middleware expects request.session to be set with a csrf_token
    attribute, e.g. session.csrf_token.
    """
    def __init__(self, signing_key=None, ttl_minutes=None):
        super().__init__()
        if not signing_key:
            signing_key = uuid4().hex
            console.warning(
                message=(
                    'no csrf signing key provided. '
                    'using random default'
                ),
                data={'random_key': signing_key}
            )

        if not isinstance(signing_key, bytes):
            signing_key = signing_key.encode('utf-8')

        self.signing_key = signing_key
        self.ttl_minutes = ttl_minutes

    @property
    def app_types(self) -> Tuple[Type['Application']]:
        return (FalconService, )

    def pre_request(
        self,
        action: 'Action',
        request: 'Request',
        raw_args: Tuple,
        raw_kwargs: Dict
    ):
        # NOTE: set csrf_protected=False in your Action decorators
        # in order to skip this middleware...
        csrf_protected = action.decorator.kwargs.get('csrf_protected')
        if csrf_protected is None:
            csrf_protected = True

        if not csrf_protected:
            return

        http_request = raw_args[0]
        token = http_request.headers.get('CSRF-TOKEN')
        session = request.session

        # bail if token missing or mismatching
        if (not (token and session)) or (session.csrf_token != token):
            raise Exception('CSRF token missing or unrecognized')

        now = datetime.now()
        iv = session.csrf_aes_cbc_iv   # cipher mode initialization vector

        # decode & decrypt the AES encrypted CSRF token
        try:
            aes_cipher = AES.new(self.signing_key, AES.MODE_CBC, iv)
            json_str = unpad(aes_cipher.decrypt(b64decode(token)), AES.block_size)
            token_components = self.app.json.decode(json_str)
            expires_at_isoformat = token_components[0]

            # convert expiration date string to a datetime object
            if expires_at_isoformat is not None:
                expires_at = datetime.fromisoformat(expires_at_isoformat)
            else:
                expires_at = None

        except Exception:
            console.error(
                message='failed to decode CSRF token',
                data={'token': token, 'session_id': session._id}
            )
            raise

        # save crsf data to request for use in post_request
        request.context.csrf = {
            'session_id': token_components[1],
            'expires_at': token_components[0],
            'csrf_token': token,
        }

        # abort request if token is expired (provided it has an expiry)
        if (expires_at is not None) and (now >= expires_at):
            console.error(
                message='received expired CSRF token',
                data={
                    'csrf_token': token,
                    'expires_at': expires_at,
                    'session_id': session._id,
                }
            )
            raise Exception('invalid CSRF token')

        if request.context.csrf['session_id'] != session._id:
            console.error(
                message='CSRF token session ID mismatch',
                data={
                    'csrf_session_id': request.context.csrf['session_id'],
                    'request_session_id': session._id,
                }
            )
            raise Exception('invalid CSRF token')
        
    def post_request(
        self,
        action: 'Action',
        request: 'Request',
        result,
    ):
        """
        Create and set a new token if the current one is expired or
        non-existent.
        """
        http_request, http_response = request.raw_args[:2]
        csrf_data = request.context.get('csrf')
        session = request.session
        regenerate_token = action.decorator.kwargs.get(
            'refresh_csrf_token', False
        )
        if (not csrf_data) or regenerate_token:
            console.debug(f'generating CSRF token for session {session._id}')
            self.create_and_set_token(http_request, http_response, session)

    def create_and_set_token(self, http_request, http_response, session):
        """
        Generate a new CSRF token, using a mix of random values and
        client-specific data.
        """

        # if there is an expiration date for the token, create the
        # datetime ISO format string for it...
        expires_at = None
        if self.ttl_minutes is not None:
            expires_at = (
                datetime.now() + timedelta(minutes=self.ttl_minutes)
            ).isoformat()

        # create the JSON array that we encrypt as the token
        json_token = self.app.json.encode([
            expires_at,
            session._id,
            # sources of entropy:
            http_request.user_agent,
            http_request.access_route[0].lower(),  # client IP
        ]).encode('utf-8')

        # encrypt and B64 encode the token
        iv = get_random_bytes(16)  # AKA "initialization vector"
        aes_cipher = AES.new(self.signing_key, AES.MODE_CBC, iv)
        encypted_token = b64encode(
            aes_cipher.encrypt(pad(json_token, AES.block_size))
        ).decode('utf-8')

        # update the user session's CSRF fields
        session.csrf_token = encypted_token
        session.csrf_aes_cbc_iv = iv

        # send CSRF token back to client
        http_response.set_header('CSRF-TOKEN', encypted_token)



from typing import Text, Tuple, List, Set, Dict, Type, Union, Callable
from collections import defaultdict

from ravel.constants import ID
from ravel.util.loggers import console
from ravel.util.misc_functions import get_class_name
from ravel.util import is_resource, is_batch
from ravel.query.order_by import OrderBy
from ravel.query.request import Request

from .resolver import Resolver


class ResolverManager(object):
    @classmethod
    def copy(cls, manager):
        copy = cls()
        copy._resolvers = manager._resolvers.copy()
        copy._tag_2_resolvers = manager._tag_2_resolvers.copy()
        copy._required_resolvers = manager._required_resolvers.copy()
        copy._private_resolvers = manager._private_resolvers.copy()
        return copy

    def __init__(self):
        self._resolvers = {}
        self._tag_2_resolvers = defaultdict(dict)
        self._required_resolvers = set()
        self._private_resolvers = set()

    def __repr__(self):
        id_resolver = self._resolvers.get(ID)
        owner_name = None
        if id_resolver.is_bootstrapped:
            owner_name = get_class_name(id_resolver.owner)
        return (
            f'{get_class_name(self)}('
            f'count={len(self)}, '
            f'owner={owner_name}'
            f')'
        )

    def __getattr__(self, tag):
        return self.by_tag(tag)

    def __getitem__(self, name):
        return self._resolvers.get(name)

    def __setitem__(self, name, resolver):
        assert name == resolver.name
        self[name] = resolver

    def __iter__(self):
        return iter(self._resolvers)

    def __contains__(self, obj):
        if isinstance(obj, Resolver):
            return obj.name in self._resolvers
        else:
            return obj in self._resolvers

    def __len__(self):
        return len(self._resolvers)

    def get(self, key, default=None):
        return self._resolvers.get(key, default)

    def keys(self):
        return list(self._resolvers.keys())

    def values(self):
        return list(self._resolvers.values())

    def items(self):
        return list(self._resolvers.items())

    @property
    def required_resolvers(self) -> Set['Resolver']:
        return self._required_resolvers

    @property
    def private_resolvers(self) -> Set['Resolver']:
        return self._private_resolvers

    def sort(self) -> List['Resolver']:
        return Resolver.sort(list(self._resolvers.values()))

    def register(self, resolver):
        name = resolver.name
        old_resolver = self._resolvers.get(name)
        if old_resolver is not None:
            del self._resolvers[name]
            if old_resolver.required:
                self._required_resolvers.remove(old_resolver)
            if old_resolver.private:
                self._private_resolvers.remove(old_resolver)
            for tag in old_resolver.tags():
                del self._tag_2_resolvers[tag][name]

        self._resolvers[name] = resolver

        if resolver.required:
            self._required_resolvers.add(resolver)
        if resolver.private:
            self._private_resolvers.add(resolver)

        for tag in resolver.tags():
            self._tag_2_resolvers[tag][name] = resolver

    def by_tag(self, tag, invert=False):
        if not invert:
            return self._tag_2_resolvers.get(tag, {})
        else:
            resolvers = {}
            keys_to_exclude = self._tag_2_resolvers.get(tag, {}).keys()
            for tag_key, resolver_dict in self._tag_2_resolvers.items():
                if tag_key != tag:
                    resolvers.update(resolver_dict)
            return resolvers



from .abstract_http_server import (
    AbstractHttpServer, Endpoint, EndpointDecorator,
    HttpClient
)
from .abstract_wsgi_service import AbstractWsgiService
from .http_server import HttpServer
from .websocket_server import WebsocketServer



from typing import Callable, Tuple, Dict

from ravel import Middleware
from ravel.app.template import JinjaTemplateRenderer


class RenderTemplate(Middleware):
    def __init__(self, renderer: 'TemplateRenderer', mime_type='text/html'):
        self._renderer = renderer
        self._content_type = mime_type

    def post_request(
        self,
        action: 'Action',
        request: 'Request',
        result,
    ):
        if action.template is not None:
            assert isinstance(action.template, str)
            assert (result is None) or isinstance(result, dict)
            tpl_filename = action.template
            http_resp = request.raw_args[1]
            http_resp.content_type = self._content_type 
            http_resp.body = self._renderer.render(tpl_filename, result)


from appyratus.utils import TimeUtils

import ravel

from ravel import (
    Application, Resource, String, Int, Email, Id,
    relationship
)


facebook = ravel.Application()


class User(Resource):
    email = Email(required=True)
    created_at = DateTime(required=True, default=True)

    @relationship(
        join=lambda: [
            (User._id, UserFriend.owner_user_id),
            (UserFriend.befriended_user_id, User._id)
        ],
        many=True
    )
    def friends(self, request):
        return request.result


class UserFriend(Resource):
    owner_user_id = Id(lambda: User, required=True)
    befriended_user_id = Id(lambda: User, required=True)





'''
class Officer(ravel.Resource):
    first_name = ravel.String()
    rank = ravel.String(default='recruit')
    ship_id = ravel.Field(private=True)
    species = ravel.Enum(
        ravel.String(),
        values=('human', 'klingon', 'vulcan'),
        nullable=False,
        default='human'
    )
    ship = ravel.Relationship(join=lambda entity: (Officer.ship_id, Ship._id))


class Ship(ravel.Resource):
    name = ravel.String()
    crew = ravel.Relationship(
        join=lambda entity: (Ship._id, Officer.ship_id),
        order_by=lambda entity: Officer.first_name.asc,
        on_add=lambda self, officer: officer.update({'ship_id': self._id}),
        on_rem=lambda self, officer: officer.update({'ship_id': None}),
        on_del=lambda self, officer: officer.update({'ship_id': None}),
        many=True
    )
    mission_names = ravel.View(load=lambda self: self.missions.name, )
    mission_count = ravel.View(
        load=lambda self: self.missions,
        transform=lambda self, data: len(data),
    )
    mission_count_with_field = ravel.View(
        load=lambda self: self.missions,
        transform=lambda self, data: str(len(data)),
        field=ravel.Int(),
    )
    missions = ravel.Relationship(
        join=lambda entity: (Ship._id, Mission.ship_id), many=True
    )


class Mission(ravel.Resource):
    name = ravel.String(nullable=False)
    description = ravel.String()
    started_at = ravel.DateTime(default=TimeUtils.utc_now, nullable=False)
    ended_at = ravel.DateTime()
    ship_id = ravel.Field()
    status = ravel.Enum(
        ravel.String(),
        values=('pending', 'active', 'closed', 'anomalous'),
        default='pending',
        nullable=False,
    )


class Planet(ravel.Resource):
    resources = fields.Nested(
        {
            'water': fields.Bool(),
            'wine': fields.List(fields.String())
        }
    )


@startrek()
def get_officer(officer: Officer) -> Officer:
    return officer


@startrek()
def get_ship(ship: Ship = None) -> Ship:
    return ship
'''



from typing import Text, Set, Dict

from appyratus.enum import EnumValueStr

from ravel.constants import (
    ID,
    REV
)

from ravel.util import is_batch, is_resource


class DumpStyle(EnumValueStr):
    @staticmethod
    def values():
        return {
            'nested',
            'side_loaded',
        }


class Dumper(object):

    @classmethod
    def get_style(cls) -> DumpStyle:
        raise NotImplementedError()

    @classmethod
    def for_style(cls, style: DumpStyle) -> 'Dumper':
        if style == DumpStyle.nested:
            return NestedDumper()
        if style == DumpStyle.side_loaded:
            return SideLoadedDumper()


class NestedDumper(Dumper):

    @classmethod
    def get_style(cls) -> DumpStyle:
        return DumpStyle.nested

    def dump(self, target: 'Resource', keys: Set = None) -> Dict:
        return self._dump_recursive(target, keys)

    def _dump_recursive(
        self, parent_resource: 'Resource', keys: Set
    ) -> Dict:

        if parent_resource is None:
            return None

        if keys:
            keys_to_dump = keys if isinstance(keys, set) else set(keys)
        else:
            keys_to_dump = parent_resource.internal.state.keys()

        record = {}
        for k in keys_to_dump:
            v = parent_resource.internal.state.get(k)
            resolver = parent_resource.ravel.resolvers.get(k)
            assert resolver is not None

            if resolver.private:
                continue

            if k in parent_resource.ravel.resolvers.relationships:
                # handle the dumping of Relationships specially
                rel = resolver
                if rel.many:
                    assert is_batch(v)
                    child_batch = v
                    record[k] = [
                        self.dump(child_resource)
                        for child_resource in child_batch
                    ]
                else:
                    if v is not None:
                        assert is_resource(v)
                    child_resource = v
                    record[k] = self.dump(child_resource)
            else:
                # dump non-Relationship state
                if k in {ID, REV}:
                    k = k[1:]
                record[k] = resolver.dump(self, v)

        return record


class SideLoadedDumper(Dumper):

    @classmethod
    def get_style(cls) -> DumpStyle:
        return DumpStyle.side_loaded

    def dump(self, target: 'Resource', keys: Set = None) -> Dict:
        links = self._dump_recursive(target)
        return {
            'target': links.pop(target._id),
            'links': links,
        }

    def _dump_recursive(
        self, parent_resource: 'Resource', links: Dict = None
    ):
        links = links if links is not None else {}
        relationships = parent_resource.ravel.resolvers.relationships

        record = {}
        for k, v in parent_resource.internal.state.items():
            if k in {ID, REV}:
                k = k[1:]

            resolver = parent_resource.ravel.resolvers[k]
            if resolver.name in relationships:
                relationship = resolver
                record[k] = getattr(v, ID)
                if k in parent_resource.internal.state:
                    self._recurse_on_entity(v, links)
            else:
                record[k] = resolver.dump(self, v)

        parent_id = getattr(parent_resource, ID)

        if parent_id not in links:
            links[parent_id] = record
        else:
            links[parent_id].update(record)

        return links

    def _recurse_on_entity(self, entity: 'Entity', links: Dict):
        if is_batch(entity):
            rel_resources = entity
        else:
            rel_resources = [entity]
        for rel_resource in rel_resources:
            self._dump_recursive(rel_resource, links)

    def _extract_relationship_state(self, resource: 'Resource') -> Dict:
        return {
            k: resource.internal.state[k]
            for k in resource.ravel.resolvers.relationships
            if k in resource.internal.state
        }



import os

# hide pygame's silly "wlcome message"
os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = 'hide'

from functools import reduce
from typing import List
from collections import defaultdict
from logging import getLogger

from appyratus.utils.sys_utils import SysUtils
from appyratus.utils.dict_utils import DictObject
from appyratus.utils.time_utils import TimeUtils

from appyratus.enum import Enum

import pygame as pg

from pygame.time import Clock

from ravel.util.misc_functions import flatten_sequence
from ravel.util.loggers import console
from ravel.app.base import Application, ActionDecorator, Action

pygame_logger = getLogger('pygame')

# Generic game components
# TODO: move to more generic location within ravel
EVENT_TYPE = Enum(
    STARTED=1,
    STOPPED=2,
    PAUSED=3,
    RESUMED=4,
    KEY_UP=5,
    KEY_DOWN=5,
    QUIT=6,
)

KEY_EVENT_TYPES = {
    pg.KEYUP,
    pg.TEXTINPUT,
    pg.KMOD_ALT,
    pg.USEREVENT + 8,
    pg.ACTIVEEVENT,
}

# Pygame-specific constants:
# TODO: move into constants.py file
GAME_STARTED = pg.USEREVENT + 0
GAME_PAUSED = pg.USEREVENT + 1
GAME_RESUMED = pg.USEREVENT + 2
GAME_STOPPED = pg.USEREVENT + 3

PYGAME_2_RAVEL_EVENT_TYPE = {
    pg.QUIT: EVENT_TYPE.QUIT,
    pg.KEYUP: EVENT_TYPE.KEY_UP,
    pg.KEYDOWN: EVENT_TYPE.KEY_DOWN,
    GAME_STARTED: EVENT_TYPE.STARTED,
    GAME_PAUSED: EVENT_TYPE.PAUSED,
    GAME_RESUMED: EVENT_TYPE.RESUMED,
    GAME_STOPPED: EVENT_TYPE.STOPPED,
}

PYGAME_WINDOW_EVENT_TYPES = {pg.QUIT}    # TODO
MOUSE_EVENT_TYPES = {pg.MOUSEBUTTONDOWN, pg.MOUSEBUTTONUP}    # TODO
PYGAME_KEY_EVENT_TYPES = {pg.KEYUP, pg.KEYDOWN}

GAME_DEFAULTS = Enum(
    FPS=34,
    WINDOW_SIZE=(1600, 600),
    TITLE="My Ravel Game",
)

# scancodes are not reliable in pygame
# http://www.asciitable.com
# https://en.wikipedia.org/wiki/ASCII#ASCII_printable_characters
# http://www.philipstorr.id.au/pcbook/book3/scancode.htm
# https://en.wikipedia.org/wiki/Keyboard_layout
# https://www.win.tue.nl/~aeb/linux/kbd/scancodes-10.html
# https://en.wikipedia.org/wiki/Scancode

GLYPH2USB_SCANCODE_KEYMAP = {
    'a': 4,
    'b': 5,
    'c': 6,
    'd': 7,
    'e': 8,
    'f': 9,
    'g': 10,
    'h': 11,
    'i': 12,
    'j': 13,
    'k': 14,
    'l': 15,
    'm': 16,
    'n': 17,
    'o': 18,
    'p': 19,
    'q': 20,
    'r': 21,
    's': 22,
    't': 23,
    'u': 24,
    'v': 25,
    'w': 26,
    'x': 27,
    'z': 29,
}

ASCII2GLYPH_KEYMAP = {
    'a': 97,
    'b': 98,
    'c': 99,
    'd': 100,
    'e': 101,
    'f': 102,
    'g': 103,
    'h': 104,
    'i': 105,
    'j': 106,
    'k': 107,
    'l': 108,
    'm': 109,
    'n': 110,
    'o': 111,
    'p': 112,
    'q': 113,
    'r': 114,
    's': 115,
    't': 116,
    'u': 117,
    'v': 118,
    'w': 119,
    'x': 120,
    'y': 121,
    'z': 122,
}

GLYPH2PYGAME_KEYMAP = {
    'a': pg.K_a,
    'b': pg.K_b,
    'c': pg.K_c,
    'd': pg.K_d,
    'e': pg.K_e,
    'f': pg.K_f,
    'g': pg.K_g,
    'h': pg.K_h,
    'i': pg.K_i,
    'j': pg.K_j,
    'k': pg.K_k,
    'l': pg.K_l,
    'm': pg.K_m,
    'n': pg.K_n,
    'o': pg.K_o,
    'p': pg.K_p,
    'q': pg.K_q,
    'r': pg.K_r,
    's': pg.K_s,
    't': pg.K_t,
    'u': pg.K_u,
    'v': pg.K_v,
    'w': pg.K_w,
    'x': pg.K_x,
    'y': pg.K_y,
    'z': pg.K_z,
}


def get_glyph_symbols(glyph):
    symbols = [glyph]
    for keymap in (GLYPH2USB_SCANCODE_KEYMAP, ASCII2GLYPH_KEYMAP, GLYPH2PYGAME_KEYMAP):
        if glyph in keymap:
            symbols.append(keymap[glyph])
    return tuple(symbols)


KEYS = Enum(
    {
        'a': get_glyph_symbols('a'),
        'b': get_glyph_symbols('b'),
        'c': get_glyph_symbols('c'),
        'd': get_glyph_symbols('d'),
        'e': get_glyph_symbols('e'),
        'f': get_glyph_symbols('f'),
        'g': get_glyph_symbols('g'),
        'h': get_glyph_symbols('h'),
        'i': get_glyph_symbols('i'),
        'j': get_glyph_symbols('j'),
        'k': get_glyph_symbols('k'),
        'l': get_glyph_symbols('l'),
        'm': get_glyph_symbols('m'),
        'n': get_glyph_symbols('n'),
        'o': get_glyph_symbols('o'),
        'p': get_glyph_symbols('p'),
        'q': get_glyph_symbols('q'),
        'r': get_glyph_symbols('r'),
        's': get_glyph_symbols('s'),
        't': get_glyph_symbols('t'),
        'u': get_glyph_symbols('u'),
        'v': get_glyph_symbols('v'),
        'w': get_glyph_symbols('w'),
        'x': get_glyph_symbols('x'),
        'y': get_glyph_symbols('y'),
        'z': get_glyph_symbols('z'),
    }
)


class GameEvent(object):
    """
    # Init process
    Event 4352 AUDIODEVICEADDED {}      
    Event 4352 AUDIODEVICEADDED {}      
    Event 32770 VIDEOEXPOSE {}          
    Event 32768 ACTIVEEVENT {'gain': 1, 'state': 1, 'window': None}
    Event 32770 VIDEOEXPOSE {}          
    Event 32770 VIDEOEXPOSE {}          
    Event 1 SCRAP_SELECTION {}          
    Event 32770 VIDEOEXPOSE {}

    # Arrow keys
    Event 768 KMOD_ALT { "key": 1073741903, "mod": 0, "scancode": 79, "unicode": "", "window": null } 
    Event 769 KEYUP { "key": 1073741903, "mod": 0, "scancode": 79, "window": null }

    # Alphanumeric keys
    Event 768 KMOD_ALT { "key": 97, "mod": 0, "scancode": 4, "unicode": "a", "window": null }           
    Event 771 TEXTINPUT { "text": "a", "window": null }
    Event 769 KEYUP { "key": 97, "mod": 0, "scancode": 4, "window": null }

    # Moving the mouse
    Event 1024 MOUSEMOTION {'pos': (699, 0), 'rel': (10, -6), 'buttons': (0, 0, 0), 'window': None}

	# Upon exit
    Event 512 WINDOWEVENT { "event": 14, "window": null }
    Event 256 QUIT

    """

    def __init__(self, event_type, **kwargs):
        self.__dict__['kwargs'] = kwargs or {}
        self.event_type = event_type
        console.debug(f'Event {event_type} mapping: "{self.get_event()}"', data=kwargs)

    def __getattr__(self, name):
        return self.__dict__['kwargs'].get(name)

    def __setattr__(self, name, value):
        self.kwargs[name] = value

    def get_event(self):
        return PYGAME_EVENT_ID_2_NAME.get(self.event_type)

    @property
    def char(self):
        try:
            text = self.unicode
            #chr(self.key)
        except ValueError as exc:
            text = ''
        return text


def get_pygame_events():
    events = {}
    for e in dir(pg):
        value = getattr(pg, e)
        if isinstance(value, int):
            events[value] = e
    return events


PYGAME_EVENT_ID_2_NAME = get_pygame_events()


def pygame_2_ravel_event(pygame_event):
    return GameEvent(pygame_event.type, **pygame_event.dict)


# ------


class PygameGame(Application):
    """
    key_event_handlers, events defined in action `key` attribute
    any_key_event_handlers, 
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.key_event_handlers = defaultdict(lambda: defaultdict(list))
        self.any_key_event_handlers = defaultdict(lambda: list())
        self.misc_event_handlers = defaultdict(lambda: list())
        self.mouse_event_handlers = defaultdict(lambda: list())
        self.user_event_handlers = defaultdict(lambda: list())
        self.window_event_handlers = defaultdict(list, {EVENT_TYPE.QUIT: [lambda event: {'is_running': False}]})
        self.local.pygame_state = DictObject()

    @property
    def state(self):
        return self.local.pygame_state

    def on_decorate(self, handler: 'Action'):
        event_type = handler.decorator.kwargs['event']

        if event_type in KEY_EVENT_TYPES:
            # look for action `key` attribute and add to "key" event handlers.
            # otherwise add to "any key" event handlers.
            key_obj = handler.decorator.kwargs.get('key')
            if key_obj is not None:
                for key in set(flatten_sequence(key_obj)):
                    self.key_event_handlers[event_type][key].append(handler)
            else:
                self.any_key_event_handlers[event_type].append(handler)

        if event_type in MOUSE_EVENT_TYPES:
            self.mouse_event_handlers[event_type].append(handler)

        if event_type >= pg.USEREVENT:
            self.user_event_handlers[event_type].append(handler)

    #
    # def on_request(
    #     self,
    #     action: 'Action',
    #     *raw_args,
    #     **raw_kwargs
    # ) -> Tuple[Tuple, Dict]:

    def on_bootstrap(
        self,
        window_size=None,
        display_mode_flags=None,
        fps=None,
        title=None,
    ):
        self.state.clock = Clock()
        self.state.is_running = False
        self.state.window_size = tuple(window_size or GAME_DEFAULTS.WINDOW_SIZE)
        self.state.fps = fps or GAME_DEFAULTS.FPS
        self.state.title = title or GAME_DEFAULTS.TITLE
        self.state.delta_t = None
        self.state.display_mode_flags = (
            reduce(lambda x, y: x | y, display_mode_flags) if display_mode_flags and len(display_mode_flags) > 1 else
            (display_mode_flags or pg.DOUBLEBUF)
        )
        pg.init()

    def on_extract(self, action, index, parameter, raw_args, raw_kwargs):
        # API methods are not called with any arguments; hence, all prepared
        # arguments are plucked from the global game state dict.
        if raw_args and index == 0 and parameter.name == 'event':
            event = raw_args[0]
            return event
        if index is not None:
            return self.state[parameter.name]
        else:
            return self.state.get(parameter.name)

    def on_start(self):
        self.state.tick = 0
        self.state.is_running = True
        self.state.screen = pg.display.set_mode(self.state.window_size, self.state.display_mode_flags)
        pg.display.set_caption(self.state.title)

        start_event = pg.event.Event(EVENT_TYPE.STARTED)
        pg.event.post(start_event)

        while self.state.is_running:
            self.state.delta_t = self.state.clock.tick(self.state.fps)
            self.state.tick += 1

            # tuple of all pressed keys by key index
            pressed = pg.key.get_pressed()

            # iterate over each pygame event
            pgevents = pg.event.get()
            for pygame_event in pgevents:
                # convert native pygame event to generic Ravel game event
                event = pygame_2_ravel_event(pygame_event)
                if not event:
                    continue
                event_type = event.event_type

                # route ravel event to appropriate handler/s
                if event_type in self.window_event_handlers:
                    handlers = self.window_event_handlers[event_type]
                    for handler in handlers:
                        fresh_state = handler(event)
                        if fresh_state:
                            self.state.update(fresh_state)

                # events that have a key attached to them
                if event_type in self.key_event_handlers and event.key:
                    key_2_handlers = self.key_event_handlers[event_type]
                    # not every event from pygame has a consistent "key" value,
                    # so we will cascade through the known locations as they
                    # are relatively exclusive to certain events.
                    # TODO this needs to be handled with better mapping of
                    # pygame events to ravel events
                    handlers = key_2_handlers[event.unicode or event.text or event.key]
                    for handler in handlers:
                        fresh_state = handler(event)
                        if fresh_state:
                            self.state.update(fresh_state)

                if event_type in self.any_key_event_handlers and event.key:
                    handlers = self.any_key_event_handlers[event_type]
                    print('found handlers', handlers)
                    print('event', event_type, event.key, event.text, event.unicode)
                    for handler in handlers:
                        fresh_state = handler(event)
                        if fresh_state:
                            self.state.update(fresh_state)

                if event_type in self.user_event_handlers:
                    handlers = self.user_event_handlers[event_type]
                    for handler in handlers:
                        fresh_state = handler(event)
                        if fresh_state:
                            self.state.update(fresh_state)

                if event_type in MOUSE_EVENT_TYPES:
                    handlers = self.mouse_event_handlers[event_type]
                    for handler in handlers:
                        fresh_state = handler(event)
                        if fresh_state:
                            self.state.update(fresh_state)

                if event_type in PYGAME_WINDOW_EVENT_TYPES:
                    handlers = self.window_event_handlers[event_type]
                    for handler in handlers:
                        fresh_state = handler(event)
                        if fresh_state:
                            self.state.update(fresh_state)

                if event_type == pg.QUIT:
                    pg.quit()

            dirty_rects = self.on_update(self.state.tick)
            self.on_draw(self.state.tick)

            # TODO: collect dirty_rects
            # TODO: update dirty_rects generated by on_update instead of flip
            pg.display.flip()

    def on_update(self, tick) -> List[pg.Rect]:
        pass

    def on_draw(self, tick):
        console.debug(f'Tick {self.state.tick}')



import traceback

from collections import defaultdict
from typing import Text, Type, Callable, List, Dict

import requests

from appyratus.utils.string_utils import StringUtils

from ravel.app.base import Application, Action, ActionDecorator
from ravel.util import get_class_name
from ravel.util.misc_functions import get_callable_name
from ravel.util.type_checking import is_resource, is_batch
from ravel.util.loggers import console

DEFAULT_HOST = 'localhost'
DEFAULT_PORT = 8081


class AbstractHttpServer(Application):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.route_2_endpoint = defaultdict(dict)
        self.host = None
        self.port = None

    @property
    def decorator_type(self) -> Type['EndpointDecorator']:
        return EndpointDecorator

    @property
    def action_type(self) -> Type['Endpoint']:
        return Endpoint

    def on_decorate(self, endpoint: 'Endpoint'):
        endpoint.routes.append(
            f'/{StringUtils.dash(endpoint.name).lower()}'
        )
        for route in endpoint.routes:
            if route not in self.route_2_endpoint:
                console.debug(f'routing {route} to {endpoint.name} action')
                self.route_2_endpoint[route][endpoint.method] = endpoint

    def on_bootstrap(self, host: Text = None, port: int = None):
        self.host = host or DEFAULT_HOST
        self.port = port or DEFAULT_PORT

    def route(self, method, route, args=None, kwargs=None):
        method = method.lower()
        route = route.lower()
        method_2_endpoint = self.route_2_endpoint.get(route)
        if method_2_endpoint:
            endpoint = method_2_endpoint.get(method)
            return endpoint(*(args or tuple()), **(kwargs or dict()))
        return None

    def build_client(self, host, port, scheme='http'):
        if not self.is_bootstrapped:
            raise Exception(f'{self} must be bootstrapped')
        return HttpClient(self, scheme, host, port)


class EndpointDecorator(ActionDecorator):
    def __init__(self,
        app: 'AbstractHttpServer',
        method: Text,
        route: Text = None,
        routes: List[Text] = None,
        path: Text = None,
        paths: List[Text] = None,
        *args, **kwargs
    ):
        super().__init__(app, *args, **kwargs)
        self.method = method.lower()

        routes = set(self._build_routes_list(route, routes))
        routes |= set(self._build_routes_list(path, paths))
        self.routes = list(routes)

    @staticmethod
    def _build_routes_list(route, routes):
        """
        Combine `route` and `routes` constructor kwargs.
        """
        route_set = set()
        if route is not None:
            if isinstance(route, (list, tuple, set)):
                route_set.update(route)
            else:
                assert isinstance(route, str)
                route_set.add(route)
        if routes:
            route_set.update(routes)
        # normalize all routes to lower case and ensure they all begin with a
        # single forward slash and no trailing slash.
        return [
            '/' + route.strip('/').lower()
            for route in route_set
        ]


class Endpoint(Action):
    """
    Stores metadata related to the "target" callable, which in the Http context
    is the action of some URL route.
    """

    def __init__(self, target, decorator):
        super().__init__(target, decorator)
        self.method = decorator.method
        self.routes = decorator.routes

    def __repr__(self):
        return '{}({})'.format(
            get_class_name(self),
            ', '.join([
                f'name={self.name}',
                f'method={self.decorator.method.upper()}',
                f'routes={self.decorator.routes}',
            ])
        )

    @classmethod
    def from_function(cls, app, func, method: str, route: str) -> 'Endpoint':
        return cls(func, EndpointDecorator(app, method=method, route=route))


class HttpClient(object):
    def __init__(
        self,
        app: 'AbstractHttpServer',
        scheme: Text,
        host: Text,
        port: int,
        headers: Dict = None,
    ):
        self._app = app
        self._handlers = {}
        self._host = host
        self._port = port
        self._scheme = scheme
        self._endpoint_name_2_route = {}
        self._response_hooks = []
        self._request_hooks = []
        self.headers = headers or {}

        for method_2_endpoint in self._app.route_2_endpoint.values():
            for http_method, endpoint in method_2_endpoint.items():
                primary_route = endpoint.routes[0]
                self._handlers[endpoint.name] = self._build_handler(
                    endpoint.name, http_method, primary_route
                )

    def __getattr__(self, route: Text) -> Callable:
        handler = self._handlers[route]
        return handler

    def _build_handler(self, name, method: Text, route: Text) -> Callable:
        self._endpoint_name_2_route[name] = route

        def handler(json=None, data=None, params=None, headers=None, path=None):

            def replace_objects_with_ids(data: Dict):
                data = data.copy()
                for k, v in list(data.items()):
                    if is_resource(v) or is_batch(v):
                        data[k] = v._id
                return data

            if isinstance(json, dict):
                json = replace_objects_with_ids(json)

            path_vars = replace_objects_with_ids(
                dict(
                    json or {},
                    **(data if isinstance(data, dict) else {}),
                    **(path or {})
                )
            )

            route = self._endpoint_name_2_route[name].format(**path_vars)
            headers = dict(self.headers, **(headers or {}))
            url = ('{}://{}:{}/' + route.strip('/')).format(
                self._scheme, self._host, self._port
            )

            kwargs = {
                'method': method,
                'url': url,
                'json': json,
                'params': params,
                'headers': headers,
            }

            for hook in self._request_hooks:
                hook(self, **kwargs)

            response = requests.request(**kwargs)

            for hook in self._response_hooks:
                hook(self, response, **kwargs)

            return response

        return handler

    def add_request_hook(self, hook):
        self._request_hooks.append(hook)

    def add_response_hook(self, hook):
        self._response_hooks.append(hook)


from typing import Union, Type
from inspect import getmembers, ismethod

from .action import Action


class ActionDecorator(object):
    def __init__(self, app: 'Application', *args, **kwargs):
        super().__init__()
        self.app = app
        self.args = args
        self.kwargs = kwargs
        self._api_object = None

    def __call__(self, obj) -> Union['Action', Type]:
        if isinstance(obj, type):
            # interpret all non-private methods as action functions
            # and register them all
            api_type = obj
            self._api_object = api_type(self.app)
            predicate = lambda x: (
                (ismethod(x) and x.__name__[0] != '_')
                or isinstance(x, Action)
            )
            for k, v in getmembers(self.api_object, predicate=predicate):
                if isinstance(v, Action):
                    # customize existing action
                    existing_action = self.app.actions.get(v.name)
                    if not existing_action:
                        # it's an action but for a different Application
                        # i.e. existing_action.app is not self.app
                        continue

                    kwargs = self.kwargs.copy()
                    kwargs.update(existing_action.decorator.kwargs)

                    new_decorator = type(self)(self.app, *self.args, **kwargs)
                    new_decorator._api_object = self._api_object
                    new_decorator.setup_action(v.target, True)
                else:
                    self.setup_action(v.__func__, False)
            return api_type
        else:
            func = obj
            action = self.setup_action(func, False)
            func._ravel_action = action
            return func

    def setup_action(self, func, overwrite):
        action = self.app.action_type(func, self)
        self.app.add_action(action, overwrite=overwrite)
        self.app.on_decorate(action)
        return action

    @property
    def api_object(self) -> 'Api':
        return self._api_object



import re
import os
import inspect
import importlib

from typing import Dict, Callable
from os.path import splitext

from appyratus.files.json import Json
from appyratus.utils.dict_utils import DictObject

from ravel.logging import ConsoleLoggerInterface

# TODO: allow package_name to refer to a module rather than a package

class Scanner:
    """
    The Scanner recursively walks the filesystem, rooted at a Python package
    or module filepath, and matches each object contained in each Python
    source file against a logical predicate. If the predicate matches, the
    object is passed into the overriden on_match instance method.
    """

    def __init__(
        self,
        predicate: Callable = None,
        callback: Callable = None,
    ):
        self.log = ConsoleLoggerInterface('scanner')
        self.context = DictObject()
        if predicate:
            self.predicate = predicate
        if callback:
            self.on_match = callback

    def scan(self, package_name, context: Dict = None, verbose=False):
        context = dict(self.context.to_dict(), **(context or {}))
        context = DictObject(context)

        root_module = importlib.import_module(package_name)
        root_filename = os.path.basename(root_module.__file__)

        if root_filename != '__init__.py':
            self.scan_module(root_module, context)
        else:
            package_dir = os.path.split(root_module.__file__)[0]
            if re.match(f'\./', package_dir):
                # ensure we use an absolute path for the package dir
                # to prevent strange string truncation results below
                package_dir = os.path.realpath(package_dir)
            package_path_len = package_name.count('.') + 1
            package_parent_dir = '/' + '/'.join(
                package_dir.strip('/').split('/')[:-package_path_len]
            )

            for dir_name, sub_dirs, file_names in os.walk(package_dir):
                file_names = set(file_names)

                if '.ravel' in file_names:
                    dot_file_path = os.path.join(dir_name, '.ravel')
                    dot_data = Json.read(dot_file_path) or {}
                    ignore = dot_data.get('scanner', {}).get('ignore', False)

                    if ignore:
                        self.log.debug(f'scanner ignoring {dir_name}')
                        sub_dirs.clear()
                        continue

                if '__init__.py' in file_names:
                    dir_name_offset = len(package_parent_dir)
                    pkg_path = dir_name[dir_name_offset + 1:].replace("/", ".")
                    for file_name in file_names:
                        if file_name.endswith('.py'):
                            mod_path = f'{pkg_path}.{splitext(file_name)[0]}'
                            try:
                                module = importlib.import_module(mod_path)
                            except Exception as exc:
                                self.on_import_error(exc, mod_path, context)
                                continue
                            self.scan_module(module, context)

        self.context = context
        return context

    def scan_module(self, module, context):
        if None in module.__dict__:
            # XXX: why is this happenings?
            del module.__dict__[None]
        for k, v in inspect.getmembers(module, predicate=self.predicate):
            # if verbose:
            #     console.debug(
            #         f'scanner matched "{k}" '
            #         f'{str(v)[:40] + "..." if len(str(v)) > 40 else v}'
            #     )
            try:
                self.on_match(k, v, context)
            except Exception as exc:
                self.on_match_error(exc, module, context, k, v)

    def predicate(self, value) -> bool:
        return True

    def on_match(self, name, value, context):
        context[name] = value

    def on_import_error(self, exc, module_path, context):
        self.log.exception(
            message='scanner encountered an import error',
            data={
                'module': module_path,
            }
        )

    def on_match_error(self, exc, module, context, name, value):
        self.log.exception(
            message=f'scanner encountered an error while scanning object',
            data={
                'module': module.__name__,
                'object': name,
                'type': type(value),
            }
        )



from appyratus.enum import EnumValueStr


class Dialect(EnumValueStr):
    @staticmethod
    def values():
        return {
            'postgresql',
            'mysql',
            'sqlite',
        }



import time

import base36

from datetime import datetime
from collections import defaultdict
from typing import Dict, List, Type, Set, Text, Tuple
from threading import local
from abc import ABCMeta, abstractmethod

from appyratus.env import Environment
from appyratus.schema.fields import UuidString
from appyratus.utils.dict_utils import DictObject
from appyratus.utils.type_utils import TypeUtils

from ravel.util.loggers import console
from ravel.util.misc_functions import get_class_name
from ravel.exceptions import RavelError
from ravel.constants import ID

from .store_history import StoreHistory, StoreEvent


class StoreError(RavelError):
    pass


class StoreMeta(ABCMeta):
    def __init__(cls, name, bases, dict_):
        ABCMeta.__init__(cls, name, bases, dict_)
        cls.ravel = DictObject()
        cls.ravel.local = local()
        cls.ravel.local.is_bootstrapped = False


class Store(object, metaclass=StoreMeta):

    env = Environment()
    ravel = None

    def __init__(self, *args, **kwargs):
        self._history = StoreHistory(store=self)
        self._is_bound = False
        self._resource_type = None

    def __repr__(self):
        if self.is_bound:
            return (
                f'<{get_class_name(self)}'
                f'(resource_type={get_class_name(self.resource_type)})>'
            )
        else:
            return (f'<{get_class_name(self)}>')

    def dispatch(
        self,
        method_name: Text,
        args: Tuple = None,
        kwargs: Dict = None
    ):
        """
        Delegate a Store call to the named method, performing any side-effects,
        like creating a StoreHistory event if need be. This is used internally
        by Resource to call into store methods.
        """
        # call the requested Store method
        func = getattr(self, method_name)
        exc = None
        try:
            result = func(*(args or tuple()), **(kwargs or {}))
        except Exception as exc:
            raise exc

        # create and store the Store call in a history event
        if self.history.is_recording_method(method_name):
            event = StoreEvent(method_name, args, kwargs, result, exc)
            self._history.append(event)

        # finally faise the exception if one was generated
        if exc is not None:
            data = {'method': method_name, 'args': args, 'kwargs': kwargs}
            if not isinstance(exc, RavelError):
                raise RavelError(data=data, wrapped_exception=exc)
            else:
                exc.data.update(data)
                raise exc

        return result

    @property
    def is_bound(self) -> bool:
        return self._is_bound

    @property
    def resource_type(self) -> Type['Resource']:
        return self._resource_type

    @property
    def schema(self) -> 'Schema':
        if self._resource_type is not None:
            return self._resource_type.ravel.schema
        return None

    @property
    def app(self) -> 'Application':
        return self.ravel.app

    @property
    def history(self) -> 'StoreHistory':
        return self._history

    def replay(
        self,
        history: StoreHistory = None,
        reads=True,
        writes=True
    ) -> Dict[StoreEvent, object]:
        """
        Replay events (interface calls) from this or another store's history,
        returning an ordered mapping from the event object to the return value
        of the corresponding store method.

        Args:
        - `history`: the history to replay in this store.
        - `reads`: replay "read" events, like query, get, get_many.
        - `writes`: replay "write" events, like create, update, etc.
        """
        results = OrderedDict()

        # use this store's own history if none specified
        if history is None:
            history = self.history

        for event in history:
            is_read = event.method in self.history.read_method_names
            is_write = event.method in self.history.write_method_names
            if (is_read and reads) or (is_write and writes):
                func = getattr(self, event.method)
                result = func(*event.args, **event.kwargs)
                results[event] = result

        return results

    def bind(self, resource_type: Type['Resource'], **kwargs):
        t1 = datetime.now()
        self._resource_type = resource_type
        self.on_bind(resource_type, **kwargs)

        t2 = datetime.now()
        secs = (t2 - t1).total_seconds()
        console.debug(
            f'bound {TypeUtils.get_class_name(resource_type)} to '
            f'{TypeUtils.get_class_name(self)} '
            f'in {secs:.2f}s'
        )

        self._is_bound = True

    @classmethod
    def bootstrap(cls, app: 'Application' = None, **kwargs):
        """
        Perform class-level initialization, like getting
        a connectio pool, for example.
        """
        t1 = datetime.now()

        cls.ravel.app = app
        cls.on_bootstrap(**kwargs)

        cls.ravel.local.is_bootstrapped = True

        t2 = datetime.now()
        secs = (t2 - t1).total_seconds()
        console.debug(
            f'bootstrapped {TypeUtils.get_class_name(cls)} '
            f'in {secs:.2f}s'
        )

        return cls

    @classmethod
    def on_bootstrap(cls, **kwargs):
        pass

    def on_bind(cls, resource_type: Type['Resource'], **kwargs):
        pass

    @classmethod
    def is_bootstrapped(cls):
        return getattr(cls.ravel.local, 'is_bootstrapped', False)

    @classmethod
    def has_transaction(cls):
        raise NotImplementedError()

    @classmethod
    def begin(cls, **kwargs):
        raise NotImplementedError()

    @classmethod
    def commit(cls, **kwargs):
        raise NotImplementedError()

    @classmethod
    def rollback(cls, **kwargs):
        raise NotImplementedError()

    def create_id(self, record: Dict) -> object:
        """
        Generate and return a new ID for the given not-yet-created record. If
        this method returns None, the backend database/storage technology
        must create and return it instead.
        """
        new_id = record.get(ID)
        if new_id is None:
            new_id = self.resource_type.ravel.defaults[ID]()
        return new_id

    def increment_rev(self, rev: Text = None) -> Text:
        time_ms = int(1000000 * time.time())

        if rev:
            rev_no = int(rev.split('-')[1], 36) + 1
        else:
            rev_no = 1

        return f'{base36.dumps(time_ms)}-{base36.dumps(rev_no)}'


    @abstractmethod
    def exists(self, _id) -> bool:
        """
        Return True if the record with the given _id exists.
        """

    @abstractmethod
    def exists_many(self, _ids: Set) -> Dict[object, bool]:
        """
        Return a mapping from _id to a boolean, indicating if the specified
        resource exists.
        """

    @abstractmethod
    def count(self) -> int:
        """
        Return the total number of stored records.
        """
    @abstractmethod
    def query(
        self,
        predicate: 'Predicate',
        fields: Set[Text] = None,
        **kwargs
    ) -> List[Dict]:
        """
        TODO: rename to "select"
        Return all records whose fields match a logical predicate.
        """

    @abstractmethod
    def fetch(self, _id, fields: Dict = None) -> Dict:
        """
        Read a single record by _id, selecting only the designated fields (or
        all by default).
        """

    @abstractmethod
    def fetch_many(self, _ids: List, fields: Dict = None) -> Dict:
        """
        Read multiple records by _id, selecting only the designated fields (or
        all by default).
        """

    @abstractmethod
    def fetch_all(self, fields: Set[Text] = None) -> Dict:
        """
        Return all records managed by this Store.
        """

    @abstractmethod
    def create(self, data: Dict) -> Dict:
        """
        Create a new record with the _id. If the _id is contained is not
        contained in the data dict nor provided as the _id argument, it is the
        responsibility of the Store class to generate the _id.
        """

    @abstractmethod
    def create_many(self, records: List[Dict]) -> List[Dict]:
        """
        Create a new record.  It is the responsibility of the Store class to
        generate the _id.
        """

    @abstractmethod
    def update(self, _id, data: Dict) -> Dict:
        """
        Update a record with the data passed in.
        """

    @abstractmethod
    def update_many(self, _ids: List, data: List[Dict] = None) -> List[Dict]:
        """
        Update multiple records. If a single data dict is passed in, then try to
        apply the same update to all records; otherwise, if a list of data dicts
        is passed in, try to zip the _ids with the data dicts and apply each
        unique update or each group of identical updates individually.
        """

    @abstractmethod
    def delete(self, _id) -> None:
        """
        Delete a single record.
        """

    @abstractmethod
    def delete_many(self, _ids: List) -> None:
        """
        Delete multiple records.
        """

    @abstractmethod
    def delete_all(self) -> None:
        """
        Delete all records.
        """



import traceback

from threading import current_thread

from ravel.util.misc_functions import get_class_name


class TransactionManager:
    def __init__(self, app: 'Application'):
        self.log = app.log
        self.store_classes = set(
            b.store_class for b in app.manifest.bindings
        )

    def __repr__(self):
        return (
            f'{get_class_name(self)}(thread={current_thread.name})'
        )

    def __enter__(self):
        self.begin()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if exc_type is None:
            self.commit()
        else:
            self.log.error(
                message=f'aborting transaction due to {exc_value}',
                data={'traceback': traceback.format_exc().split('\n')}
            )
            self.rollback()

    def begin(self, **kwargs):
        for store_class in self.store_classes:
            try:
                store_class.begin(**kwargs)
            except Exception as exc:
                self.log.error(
                    message=f'failed to create transaction',
                    data={'store': store_class}
                )
                raise

    def commit(self, **kwargs):
        for store_class in self.store_classes:
            try:
                exc = None
                if store_class.has_transaction():
                    store_class.commit(**kwargs)
            except Exception as exc:
                self.log.error(
                    message=f'failed to commit transaction. rolling back',
                    data={'store': store_class}
                )
            if exc:
                try:
                    store_class.rollback()
                except Exception:
                    self.log.exception(
                        message=f'failed to rollback transaction',
                        data={'store': store_class}
                    )
        if exc:
            raise Exception('a transaction failed')

    def rollback(self, **kwargs):
        for store_class in self.store_classes:
            try:
                if store_class.has_transaction():
                    store_class.rollback(**kwargs)
            except Exception as exc:
                self.log.error(
                    message=f'failed to rollback transaction',
                    data={'store': store_class}
                )
                raise



from typing import Text

from ravel.util.misc_functions import get_class_name


class ParameterAssignment(object):
    """
    This is an internal data structure, used to facilitate the syntactic sugar
    that allows you to write to query.params via funcion call notation, like
    query.foo('bar') instead of query.params['foo'] = bar.

    Instances of this class just store the query whose parameter we are going to
    set and the name of the dict key or "param name". When called, it writes the
    single argument supplied in the call to the params dict of the query.
    """

    def __init__(self, owner, name: Text):
        self.owner = owner
        self.name = name

    def __call__(self, value):
        """
        Store the `param` value in the Query's parameters dict.
        """
        self.owner.parameters[self.name] = value
        return self.owner

    def __repr__(self):
        return (
            f'{get_class_name(self)}('
            f'parameter={self.name}'
            f')'
        )



import traceback


class RavelError(Exception):
    def __init__(
        self,
        message: str,
        data: dict = None,
        wrapped_exception: Exception = None,
        logged_traceback_depth: int = None
    ):
        super().__init__(message)
        self.data = data or {}
        self.wrapped_exception = wrapped_exception
        self.logged_traceback_depth = logged_traceback_depth
        self.wrapped_traceback = None
        if wrapped_exception:
            self.wrapped_traceback = traceback.format_exc()

    def to_dict(self):
        return {
            'message': str(self),
            'data': self.data,
        }


class RelationshipError(RavelError):
    pass


class NotFound(RavelError):
    pass


class NotAuthorized(RavelError):
    pass


class ManifestError(RavelError):
    pass


class ValidationError(RavelError):
    pass


class RelationshipArgumentError(RavelError):
    pass


class ResourceError(RavelError):
    pass

class NotBootstrapped(RavelError):
    pass



from typing import Text
from base64 import b64decode, b64encode

from Crypto.Cipher import AES as AES_Cipher
from Crypto.Random import get_random_bytes
from Crypto.Util.Padding import pad, unpad

from ravel.util.uuid_util import random_uuid
from ravel.util.json_encoder import JsonEncoder


class AES:
    def __init__(self, key: Text = None, iv: bytes = None):
        self.json = JsonEncoder()
        self.iv = iv or get_random_bytes(16)
        self.key = (key or random_uuid().hex).encode('utf-8')

    def encrypt_CBC(self, value: object) -> Text:
        json_str = self.json.encode(value).encode('utf-8')
        cipher = AES_Cipher.new(self.key, AES_Cipher.MODE_CBC, self.iv)
        enc_bytes = cipher.encrypt(pad(json_str, AES_Cipher.block_size))
        return b64encode(enc_bytes).decode('utf-8')

    def decrypt_CBC(self, encrypted_value: Text) -> object:
        cipher = AES_Cipher.new(self.key, AES_Cipher.MODE_CBC, self.iv)
        enc_bytes = b64decode(encrypted_value.encode('utf-8'))
        json_str = unpad(cipher.decrypt(enc_bytes), AES_Cipher.block_size)
        return self.json.decode(json_str)


from typing import Text, Dict, List, Union
from datetime import datetime, date

import numpy as np

from ravel.util.misc_functions import get_class_name, normalize_to_tuple

MAX_UNICODE_VALUE = 0x10FFFF


class OrderBy(object):
    def __init__(self, key: Text, desc=False):
        self.key = key
        self.desc = desc

    def __repr__(self):
        return (
            f'{get_class_name(self)}('
            f'{self.key} {"desc" if self.desc else "asc"}'
            f')'
        )

    def to_sql(self) -> Text:
        return f'{self.key} {"DESC" if self.desc else "ASC"}'

    def dump(self):
        return {'key': self.key, 'desc': self.desc}

    @classmethod
    def load(cls, data: Dict) -> 'OrderBy':
        return cls(data['key'], data['desc'])

    @property
    def asc(self) -> bool:
        return not self.desc

    @staticmethod
    def sort(
        resources: List['Resource'],
        order_by: Union['OrderBy', List['OrderBy']]
    ) -> List['Resource']:
        """
        Perform a multi-key sort on the given resource list. This procedure
        approximately O(N log N).
        """
        order_by = normalize_to_tuple(order_by)

        # if we only have one key to sort by, skip the fancy indexing logic
        # below and just use built-in sorted method as nature intended.
        if len(order_by) == 1:
            key = order_by[0].key
            reverse = order_by[0].desc
            return sorted(resources, key=lambda x: x[key], reverse=reverse)

        # create functions for converting types that are not inherently
        # sortable to an integer value (which is sortable)
        converters = {
            datetime: lambda x: x.timestamp(),
            date: lambda x: x.toordinal(),
            bytes: lambda x: int.from_bytes(x, byteorder='big'),
        }

        # pre-compute the "index" keys by which the resources shall be sorted.
        # Each index is an array of ints.
        indexes = {}
        for resource in resources:
            index = []
            for x in order_by:
                value = resource.internal.state.get(x.key)
                if value is None:
                    value = 0
                if x.desc:
                    if isinstance(value, str):
                        ords = np.fromiter((ord(c) for c in value), dtype=int)
                        inverse_ords = MAX_UNICODE_VALUE - ords
                        inverse_value = ''.join(chr(n) for n in inverse_ords)
                        index.append(inverse_value)
                    else:
                        convert = converters.get(type(value))
                        if convert:
                            value = convert(value)
                        index.append(-1 * value)
                else:
                    index.append(value)

            indexes[resource] = tuple(index)

        return sorted(resources, key=lambda x: indexes[x])



import os

from typing import Text

from ravel import Resource, fields

BCRYPT_ROUNDS = int(os.environ.get('USER_PASSWORD_BCRYPT_ROUNDS', 14))


class User(Resource):
    email = fields.String(required=True, nullable=True, default=lambda: None)
    password = fields.BcryptString(
        rounds=BCRYPT_ROUNDS, private=True, required=True, nullable=False
    )
    phone = fields.String(nullable=True)

    @classmethod
    def __abstract__(cls):
        return True



from typing import Type, Dict, List, Set, Text
from collections import defaultdict
from copy import deepcopy

from redis import StrictRedis
from appyratus.utils.string_utils import StringUtils

from ravel.store import Store
from ravel.schema import fields
from ravel.constants import ID, REV
from ravel.util.json_encoder import JsonEncoder
from ravel.query.predicate import (
    Predicate,
    ConditionalPredicate,
    BooleanPredicate,
    OP_CODE,
)

from .redis_types import RedisClient, HashSet, StringIndex, NumericIndex


class RedisStore(Store):

    redis = None
    host = 'localhost'
    port = 6379
    db = 0

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.encoder = JsonEncoder()
        self.field_2_index_type = {
            fields.String: StringIndex,
            fields.Int: NumericIndex,
            fields.Float: NumericIndex,
            fields.DateTime: NumericIndex,
            fields.Timestamp: NumericIndex,
            fields.Uuid: NumericIndex,
            fields.Bool: NumericIndex,
        }

    @classmethod
    def on_bootstrap(cls, host=None, port=None, db=None):
        cls.host = host or cls.env.REDIS_HOST or cls.host
        cls.port = port or cls.env.REDIS_PORT or cls.port
        cls.db = db if db is not None else (cls.env.REDIS_DB or cls.db)
        cls.redis = RedisClient(host=cls.host, port=cls.port, db=cls.db)

    def on_bind(self, resource_type: Type['Resource'], **kwargs):
        self.type_name = StringUtils.snake(resource_type.__name__).lower()
        self.records = HashSet(self.redis, self.type_name)
        self.revs = HashSet(self.redis, f'{self.type_name}_revisions')
        self.indexes = self._bind_indexes()

    def _bind_indexes(self):
        indexes = {}
        for k, field in self.resource_type.Schema.fields.items():
            index_name = f'{self.type_name}:{k.lower()}'
            index_type = self.field_2_index_type.get(field.__class__)
            if index_type is None:
                index_type = StringIndex
            indexes[k] = index_type(self.redis, index_name)
        return indexes

    def exists(self, _id) -> bool:
        return (_id in self.records)

    def exists_many(self, _ids: Set) -> Dict[object, bool]:
        return {
            _id: (_id in self.records)
            for _id in _ids
        }

    def count(self) -> int:
        return len(self.records)

    def fetch(self, _id, fields: Set[Text] = None) -> Dict:
        fields = fields if isinstance(fields, set) else set(fields or [])

        pipe = self.redis.pipeline()

        self.records.get(_id)
        self.revs.get(_id)

        record_json, rev_str = pipe.execute()

        if not record_json:
            return None

        full_record = JsonEncoder.decode(record_json)
        full_record[REV] = int(rev_str) - 1

        if fields:
            return {k: full_record.get(k) for k in fields}
        else:
            return full_record

    def fetch_many(self, _ids: List, fields: Set[Text] = None) -> Dict:
        pipe = self.redis.pipeline()

        self.records.get_many(_ids, pipe=pipe)
        self.revs.get_many(_ids, pipe=pipe)

        json_records, rev_strs = pipe.execute()

        fields = fields if isinstance(fields, set) else set(fields or [])
        records = {}

        if fields:
            for record_json, _rev in zip(json_records, rev_strs):
                record = JsonEncoder.decode(record_json)
                record = {k: record.get(k) for k in fields}
                record[REV] = int(_rev) - 1
                records[record[ID]] = record
        else:
            for record_json, _rev in zip(json_records, rev_strs):
                record = JsonEncoder.decode(record_json)
                record[REV] = int(_rev) - 1
                records[record[ID]] = record

        return records

    def fetch_all(self, fields: Set[Text] = None) -> Dict:
        return self.fetch_many(list(self.records.keys()), fields=fields)

    def upsert(self, record: Dict, pipe=None) -> Dict:
        is_creating = record.get(ID) is None
        _id = self.create_id(record)

        # prepare the record for upsert
        upserted_record = record.copy()
        upserted_record[ID] = _id

        # json encode and store record JSON
        self.records[_id] = self.encoder.encode(upserted_record)

        # update indexes
        for k, v in record.items():
            self.indexes[k].upsert(_id, v)

        # add rev to record AFTER insert to avoid storing _rev in records hset
        if is_creating:
            upserted_record[REV] = 0
        else:
            upserted_record[REV] = self.revs.increment(_id)

        return upserted_record

    def create(self, data: Dict) -> Dict:
        pipe = self.redis.pipeline()
        created_record = self.upsert(data)
        pipe.execute()
        return created_record

    def update(self, _id, record: Dict) -> Dict:
        pipe = self.redis.pipeline()
        updated_record = self.upsert(record, pipe=pipe)
        pipe.execute()
        return updated_record

    def create_many(self, records: List[Dict]) -> Dict:
        pipe = self.redis.pipeline()
        created_records = [self.upsert(rec, pipe=pipe) for rec in records]
        pipe.execute()
        return created_records

    def update_many(self, _ids: List, data: Dict = None) -> Dict:
        pipe = self.redis.pipeline()
        updated_records = [self.upsert(rec, pipe=pipe) for rec in records]
        pipe.execute()
        return updated_records

    def delete(self, _id) -> None:
        pipe = self.redis.pipeline()
        if self.records.delete(_id, pipe=pipe):
            for index in self.indexes.values():
                index.delete(_id, pipe=pipe)
        pipe.execute()

    def delete_many(self, _ids: List) -> None:
        pipe = self.redis.pipeline()
        if self.records.delete_many(_ids, pipe=pipe):
            self.revs.delete_many(_ids)
            for index in self.indexes.values():
                index.delete_many(_ids, pipe=pipe)
        pipe.execute()

    def delete_all(self):
        if self.count():
            self.delete_many(self.records.keys())

    def query(self, predicate: 'Predicate', **kwargs):
        if predicate is None:
            _ids = self.records.keys()
        else:
            _ids = self.query_ids(predicate)

        records = []

        if _ids:
            _id_field = self.resource_type.Schema.fields[ID]
            json_records = self.records.get_many(_ids)
            rev_strs = self.revs.get_many(_ids)
            for json_record, rev_str in zip(json_records, rev_strs):
                record = JsonEncoder.decode(json_record)
                record[REV] = int(rev_str.decode())
                records.append(record)

        return records

    def query_ids(self, predicate, pipeline=None):
        # TODO: use a redis pipeline object

        if isinstance(predicate, ConditionalPredicate):
            index = self.indexes.get(predicate.field.name)
            if index is None:
                # TODO: raise custom exception
                raise Exception('no index')

            ids = set()
            if predicate.op == OP_CODE.EQ:
                ids = set(index.search(
                    upper=predicate.value,
                    lower=predicate.value,
                    include_upper=True,
                    include_lower=True
                ))
            elif predicate.op == OP_CODE.NEQ:
                ids = set(index.search(
                    upper=predicate.value,
                    include_upper=False,
                    include_lower=True
                ))
                ids |= set(index.search(
                    lower=predicate.value,
                    include_upper=True,
                    include_lower=False
                ))
            elif predicate.op == OP_CODE.GEQ:
                ids = set(index.search(
                    lower=predicate.value,
                    include_lower=True
                ))
            elif predicate.op == OP_CODE.LEQ:
                ids = set(index.search(
                    upper=predicate.value,
                    include_upper=True
                ))
            elif predicate.op == OP_CODE.LT:
                ids = set(index.search(
                    upper=predicate.value,
                    include_upper=False
                ))
            elif predicate.op == OP_CODE.GT:
                ids = set(index.search(
                    lower=predicate.value,
                    include_lower=False
                ))
            elif predicate.op == OP_CODE.INCLUDING:
                ids = set()
                for v in predicate.value:
                    ids |= set(index.search(
                        upper=v,
                        lower=v,
                        include_upper=True,
                        include_lower=True
                    ))
            elif predicate.op == OP_CODE.EXCLUDING:
                ids = set()
                for v in predicate.value:
                    ids |= set(index.search(
                        upper=v,
                        include_upper=False,
                        include_lower=True
                    ))
                    ids |= set(index.search(
                        lower=v,
                        include_upper=True,
                        include_lower=False
                    ))
            else:
                raise ValueError('unrecognized op: {}'.format(predicate.op))
        elif isinstance(predicate, BooleanPredicate):
            ids_lhs = self.query_ids(predicate.lhs)
            if ids_lhs:
                ids_rhs = self.query_ids(predicate.rhs)
            else:
                ids_rhs = set()
            if predicate.op == OP_CODE.OR:
                ids = ids_lhs | ids_rhs
            elif predicate.op == OP_CODE.AND:
                ids = ids_lhs & ids_rhs
            else:
                raise ValueError('unrecognized op: {}'.format(predicate.op))

        return ids



import re
import pickle
import codecs

from functools import reduce
from collections import defaultdict
from typing import Dict, Set, Text, List, Type, Tuple
from threading import local

import sqlparse

from sqlparse.sql import (
    Statement, Parenthesis, Token, Comparison,
    Identifier,
)

from appyratus.enum import Enum
from appyratus.utils.dict_utils import DictObject

from ravel.util.misc_functions import (
    flatten_sequence, is_sequence, get_class_name
)
from ravel.schema import Field, fields
from ravel.constants import ID, REV, OP_CODE
from ravel.util import is_resource, is_batch


PREDICATE_TYPE = Enum(
    BOOLEAN=1,
    CONDITIONAL=2,
)

OP_CODE_2_DISPLAY_STRING = {
    OP_CODE.EQ: '==',
    OP_CODE.NEQ: '!=',
    OP_CODE.GT: '>',
    OP_CODE.LT: '<',
    OP_CODE.GEQ: '>=',
    OP_CODE.LEQ: '<=',
    OP_CODE.INCLUDING: 'in',
    OP_CODE.EXCLUDING: 'not in',
    OP_CODE.AND: '&&',
    OP_CODE.OR: '||',
}

DISPLAY_STRING_2_OP_CODE = {
    v: k for k, v in OP_CODE_2_DISPLAY_STRING.items()
}

NON_SCALAR_OP_CODES = {
    OP_CODE.INCLUDING,
    OP_CODE.EXCLUDING,
}

# regular expressions
RE_INT = re.compile(r'\d+')
RE_FLOAT = re.compile(r'\d*(\.\d+)')
RE_STRING = re.compile(r'\'.+\'')

AND_FUNC = lambda x, y: x & y
OR_FUNC = lambda x, y: x | y


class Predicate(object):
    TYPE = PREDICATE_TYPE

    def __init__(self, code):
        self.code = code
        self.targets = set()

    def serialize(self) -> Text:
        """
        Return the Predicate as a base64 encoded pickle. This is used, for
        instance, by ravel gRPC instrumentation, for passing these objects over
        the line.
        """
        return codecs.encode(pickle.dumps(self), "base64").decode()

    @staticmethod
    def deserialize(obj) -> 'Predicate':
        """
        Return the Predicate from a base64 encoded pickle. This is used, for
        instance, by ravel gRPC instrumentation, for passing these objects over
        the line.
        """
        if isinstance(obj, str):
            return pickle.loads(codecs.decode(obj.encode(), 'base64'))
        elif isinstance(obj, Predicate):
            return obj
        else:
            raise ValueError(str(obj))

    def dump(self):
        raise NotImplementedError()

    @classmethod
    def load(cls, resource_type: Type['Resource'], data: Dict):
        if data['code'] == PREDICATE_TYPE.CONDITIONAL:
            return ConditionalPredicate.load(resource_type, data)
        elif data['code'] == PREDICATE_TYPE.BOOLEAN:
            return BooleanPredicate.load(resource_type, data)

    @classmethod
    def reduce_and(cls, *predicates) -> 'Predicate':
        return cls._reduce(AND_FUNC, predicates)

    @classmethod
    def reduce_or(cls, *predicates) -> 'Predicate':
        return cls._reduce(OR_FUNC, predicates)

    @staticmethod
    def _reduce(func, predicates: List['Predicate']) -> 'Predicate':
        predicates = flatten_sequence(p for p in predicates if p is not None)
        if not predicates:
            return None
        if len(predicates) == 1:
            return list(predicates)[0]
        else:
            return reduce(func, predicates)

    @property
    def is_conditional_predicate(self):
        return self.code == PREDICATE_TYPE.CONDITIONAL

    @property
    def is_boolean_predicate(self):
        return self.code == PREDICATE_TYPE.BOOLEAN

    def satisfy(self) -> Dict:

        def group_by_conditionals(predicate, groups=None):
            groups = defaultdict(list) if groups is None else groups
            if predicate.is_boolean_predicate:
                group_by_conditionals(predicate.lhs, groups=groups)
                group_by_conditionals(predicate.rhs, groups=groups)
            else:
                assert predicate.is_conditional_predicate
                groups[predicate.field].append(predicate)
            return groups

        groups = group_by_conditionals(self)
        values = {}
        nil = {}

        for field, group in groups.items():
            exact_value = nil
            including = set()
            excluding = set()
            lower_value = None
            lower_inclusive = None
            upper_value = None
            upper_inclusive = None

            for pred in group:
                if pred.op == OP_CODE.EQ:
                    exact_value = pred.value
                elif pred.op == OP_CODE.NEQ:
                    excluding.append(pred.op)
                elif pred.op == OP_CODE.LT:
                    if upper_value is None or pred.value <= upper_value:
                        upper_value = pred.value
                        upper_inclusive = False
                elif pred.op == OP_CODE.LEQ:
                    if upper_value is None or pred.value <= upper_value:
                        upper_value = pred.value
                        upper_inclusive = True
                elif pred.op == OP_CODE.GT:
                    if lower_value is None or pred.value >= lower_value:
                        lower_value = pred.value
                        lower_inclusive = True
                elif pred.op == OP_CODE.LEQ:
                    if lower_value is None or pred.value >= lower_value:
                        lower_value = pred.value
                        lower_inclusive = False
                elif pred.op == OP_CODE.INCLUDING:
                    including.update(pred.op_value)
                elif pred.op == OP_CODE.EXCLUDING:
                    excluding.update(pred.op_value)

            if exact_value is not nil:
                values[field.name] = exact_value
                continue

            if including:
                if excluding:
                    including -= excluding
                if including:
                    values[field.name] = random.choice(list(including))
                    continue

            bounds = None
            if lower is not None or upper is not None:
                bounds = Field.Bounds(
                    lower=lower_value,
                    upper=upper_value,
                    lower_inclusive=lower_inclusive,
                    upper_inclusive=upper_inclusive,
                    exclude=excluding,
                )

            values[field.name] = field.generate(bounds=bounds)

        return values


class ConditionalPredicate(Predicate):
    """
    A `ConditionalPredicate` specifies a comparison between the name of a field
    and a value. The field name is made available through the FieldProperty
    objects that instantiated the predicate.
    """

    def __init__(
        self,
        op: Text,
        prop: 'FieldProperty',
        value,
        ignore_field_adapter=False
    ):
        super().__init__(code=PREDICATE_TYPE.CONDITIONAL)
        self.op = op
        self.prop = prop
        self.value = value
        self.targets.add(self.prop.resolver.owner)
        self.is_scalar = op not in NON_SCALAR_OP_CODES
        self.ignore_field_adapter = ignore_field_adapter

    def __repr__(self):
        return f'{get_class_name(self)}({str(self)[1:-1]})'

    def __str__(self):
        if self.prop:
            res_class_name = get_class_name(self.prop.resolver.owner)
            lhs = f'{res_class_name}.{self.prop.resolver.field.name}'
        else:
            lhs = '[NULL]'

        val = self.value

        return f'({lhs} {OP_CODE_2_DISPLAY_STRING[self.op]} {val})'

    def __or__(self, other):
        return BooleanPredicate(OP_CODE.OR, self, other)

    def __and__(self, other):
        return BooleanPredicate(OP_CODE.AND, self, other)

    @classmethod
    def load(cls, resource_type: Type['Resource'], data: Dict):
        field_prop = getattr(resource_type, data['field'])
        return cls(data['op'], field_prop, data['value'])

    @property
    def field(self):
        return self.prop.resolver.field

    def dump(self):
        return {
            'op': self.op,
            'field': self.prop.resolver.field.name,
            'value': self.value,
            'code': self.code,
        }


class BooleanPredicate(Predicate):
    """
    A `BooleanPredicate` is used for combining children `Predicate` objects in
    a boolean expression, like (User._name == 'foo') & (User.smell == 'stink').
    LSH, and LHS stand for "left-hand side" and "right-hand side", respectively.
    """
    def __init__(self, op, lhs: 'Predicate', rhs: 'Predicate' = None):
        super().__init__(code=PREDICATE_TYPE.BOOLEAN)
        self.op = op
        self.lhs = lhs
        self.rhs = rhs

    def __or__(self, other):
        return BooleanPredicate(OP_CODE.OR, self, other)

    def __and__(self, other):
        return BooleanPredicate(OP_CODE.AND, self, other)

    def __str__(self):
        return self._build_string(self)

    def __repr__(self):
        return f'{get_class_name(self)}({self.lhs} {self.op} {self.rhs})'

    def _build_string(self, p, depth=0):
        substr = ''

        if isinstance(p.lhs, BooleanPredicate):
            substr += self._build_string(p.lhs, depth+1)
        else:
            substr += str(p.lhs)

        substr += f' {OP_CODE_2_DISPLAY_STRING[p.op]} '

        if isinstance(p.rhs, BooleanPredicate):
            substr += self._build_string(p.rhs, depth+1)
        else:
            substr += str(p.rhs)

        return f'({substr})'

    @classmethod
    def load(cls, resource_type: Type['Resource'], data: Dict):
        return cls(
            data['op'],
            Predicate.load(resource_type, data['lhs']),
            Predicate.load(resource_type, data['rhs']),
        )

    def dump(self):
        return {
            'op': self.op,
            'lhs': self.lhs.dump(),
            'rhs': self.rhs.dump(),
            'code': self.code,
        }


class PredicateParser(object):

    ravel_field_name_transform_inversions = {
        'id': ID,
        'rev': REV,
    }

    class Operand(object):
        def __init__(self, op_code, arity):
            self.op_code = op_code
            self.arity = arity

    def __init__(self, resource_type):
        self._resource_type = resource_type

    def parse(self, source: Text):
        source = f'({source})'
        stmt = sqlparse.parse(source)[0]
        predicate = self._parse_predicate(stmt[0])
        return predicate

    def _parse_predicate(self, paren: Parenthesis):
        predicate_stack = []
        op_stack = []
        for token in paren:
            if isinstance(token, Parenthesis):
                predicate = self._parse_predicate(token)
                predicate_stack.append(predicate)
                if op_stack:
                    op = op_stack[-1]
                    if len(predicate_stack) >= op.arity:
                        op_stack.pop()
                        args = [predicate_stack.pop() for i in range(op.arity)]
                        if op.op_code == 'and':
                            predicate_stack.append(Predicate.reduce_and(*args))
                        elif op.op_code == 'or':
                            predicate_stack.append(Predicate.reduce_or(*args))
                        else:
                            raise Exception()
            elif token.is_keyword:
                value = token.value.lower()
                if value in {'and', 'or'}:
                    op_stack.append(self.Operand(value, 2))
                elif value == 'in':
                    predicate_stack.append(self._parse_in_predicate(paren))
            elif isinstance(token, Comparison):
                return self._parse_comparison(token)

        return predicate_stack[0] if predicate_stack else None

    def _parse_comparison(self, comp: Comparison):
        if isinstance(comp.left, Identifier):
            ident = comp.left.value
            target = comp.right.value
        else:
            ident = comp.right.value
            target = comp.left.value

        target = target.strip("'")
        ident = self.ravel_field_name_transform_inversions.get(ident, ident)
        prop = getattr(self._resource_type, ident)

        op_code = None
        for token in comp:
            if token.ttype == sqlparse.tokens.Comparison:
                op_code = token.value
                break

        if op_code == '=':
            return prop == target
        if op_code == '!=':
            return prop != target
        if op_code == '<':
            return prop < target
        if op_code == '>':
            return prop > target
        if op_code == '>=':
            return prop >= target
        if op_code == '<=':
            return prop <= target

        raise Exception()

    def _parse_in_predicate(self, paren: Parenthesis):
        ident = None
        value_list = []
        is_negative = False
        for token in paren:
            if token.is_keyword and token.value.lower() == 'not':
                is_negative = not is_negative
            if isinstance(token, Identifier):
                if ident is None:
                    ident = token.value
                else:
                    ident = self.ravel_field_name_transform_inversions.get(
                        ident, ident
                    )
                    prop = getattr(self._resource_type, ident)
                    value_list = [
                        prop.resolver.field.process(x.strip("'").strip())[0]
                        for x in token.value[1:-1].split(',')
                    ]
                    if is_negative:
                        return prop.excluding(value_list)
                    else:
                        return prop.including(value_list)



import codecs
import pickle

import grpc

from typing import Text

from appyratus.utils.string_utils import StringUtils

from ravel.util.loggers import console
from ravel.schema import fields, Schema, Field


class GrpcClient(object):
    def __init__(self, app: 'GrpcApplication'):
        assert app.is_bootstrapped

        self._address = app.grpc.options.client_address
        self._app = app

        if app.grpc.options.secure_channel:
            self._channel = grpc.secure_channel(
                self._address, grpc.ssl_channel_credentials()
            )
        else:
            self._channel = grpc.insecure_channel(self._address)

        console.info(
            message='gRPC client initialized',
            data={
                'address': self._address,
                'secure': app.grpc.options.secure_channel
            }
        )

        GrpcApplicationStub = app.grpc.pb2_grpc.GrpcApplicationStub

        self._grpc_stub = GrpcApplicationStub(self._channel)
        self._funcs = {
            k: self._build_func(p)
            for k, p in app.actions.items()
        }

    def __getattr__(self, func_name: Text):
        return self._funcs[func_name]

    def _build_func(self, action):
        key = StringUtils.camel(action.name)
        request_type = getattr(self._app.grpc.pb2, f'{key}Request')
        send_request = getattr(self._grpc_stub, action.name)

        def func(**kwargs):
            # prepare and send the request
            request = request_type(**kwargs)
            response = send_request(request)
            # translate the native proto response message to a plain dict
            if action.streams_response:
                data = [
                    self._extract_fields(x, action.schemas.response)
                    for x in response
                ]
            else:
                data = self._extract_fields(response, action.schemas.response)
            return data

        return func

    def _extract_fields(self, message, schema):
        result = {}
        for field_name, field in schema.fields.items():
            value = getattr(message, field_name, None)
            if isinstance(field, fields.Dict):
                decoded_value = codecs.decode(value, 'base64')
                result[field_name] = pickle.loads(decoded_value)
            elif isinstance(field, fields.Nested):
                result[field_name] = self._extract_fields(value, field.schema)
            elif isinstance(field, fields.List):
                if isinstance(field.nested, Schema):
                    result[field_name] = [
                        self._extract_fields(v, field.nested) for v in value
                    ]
                elif isinstance(field.nested, fields.Dict):
                    result[field_name] = [
                        pickle.loads(codecs.decode(v, 'base64')) for v in value
                    ]
                else:
                    result[field_name] = list(value)
            elif isinstance(field, fields.Set):
                if isinstance(field.nested, Schema):
                    result[field_name] = {
                        self._extract_fields(v, field.nested)
                        for v in value
                    }
                elif isinstance(field.nested, fields.Dict):
                    result[field_name] = {
                        pickle.loads(codecs.decode(v, 'base64'))
                        for v in value
                    }
                else:
                    result[field_name] = set(value)
            elif isinstance(field, Schema):
                result[field_name] = self._extract_fields(value, field)
            else:
                result[field_name] = value
        return result



import numpy as np

from ravel.schema import fields


class Array(fields.List):
    def process(self, obj):
        processed_obj, error = super().process(obj)
        if error:
            return (None, error)
        arr = np.array(processed_obj, dtype=self.nested.np_dtype)
        return (arr, None)


import inspect 

from typing import Text, List, Set, Dict, ByteString

from appyratus.schema import Schema, fields as field_types
from appyratus.utils.dict_utils import DictObject
from appyratus.utils.string_utils import StringUtils

from ravel.app import Action
from ravel.util.misc_functions import extract_res_info_from_annotation
from ravel.util import is_resource_type, get_class_name
from ravel.util.loggers import console
from ravel.constants import ID, REV

from .util import get_stripped_schema_name
from .proto import MessageGenerator


class GrpcMethod(Action):

    _py_type_2_field_type = {
        int: field_types.Int,
        float: field_types.Float,
        str: field_types.String,
        Text: field_types.String,
        bool: field_types.Bool,
        bytes: field_types.Bytes,
        ByteString: field_types.Bytes,
        dict: field_types.Dict,
        Dict: field_types.Dict,
        list: field_types.List,
        List: field_types.List,
        set: field_types.Set,
        Set: field_types.Set,
    }

    def __init__(self, target, decorator):
        super().__init__(target, decorator)
        self._msg_gen = MessageGenerator()
        self._returns_stream = False
        self.schemas = DictObject()

    def __call__(self, *raw_args, **raw_kwargs):
        return super().__call__(*(raw_args[:1]), **raw_kwargs)

    def on_bootstrap(self):
        self._msg_name_prefix = StringUtils.camel(self.name)
        self.schemas.request = self._build_request_schema()
        self.schemas.response = self._build_response_schema()

    @property
    def streams_response(self) -> bool:
        return self._returns_stream

    def generate_request_message_type(self) -> Text:
        text = self._msg_gen.emit(self.app, self.schemas.request)
        return text + '\n' if text is not None else None

    def generate_response_message_type(self) -> Text:
        text = self._msg_gen.emit(self.app, self.schemas.response)
        return text + '\n' if text is not None else None

    def generate_protobuf_function_declaration(self, stream=None) -> Text:
        request_type_name = get_stripped_schema_name(self.schemas.request)
        response_type_name = get_stripped_schema_name(self.schemas.response)

        if stream is not None:
            stream_str = 'stream ' if stream else ''
        elif self._returns_stream:
            stream_str = 'stream '
        else:
            stream_str = ''

        return (
            f'rpc {self.name}({request_type_name}) '
            f'returns ({stream_str}{response_type_name})'
            f' {{}}'
        )

    def _build_response_schema(self):
        type_name = f'{StringUtils.camel(self.name)}ResponseSchema'
        obj = self.decorator.kwargs.get('response')

        if isinstance(obj, dict):
            schema = Schema.factory(type_name, obj)()
        elif isinstance(obj, Schema):
            response_schema_type = type(type_name, (type(obj), ), {})
            schema = response_schema_type()
        elif is_resource_type(obj):
            response_schema_type = type(type_name, (obj.Schema, ), {})
            schema = response_schema_type()
        else:
            many, resource_type_name = extract_res_info_from_annotation(
                self.signature.return_annotation
            )
            self._returns_stream = many
            if resource_type_name in self.app.manifest.resource_classes:
                base_schema_type = self.app[resource_type_name].Schema
                response_schema_type = type(type_name, (base_schema_type, ), {})
                schema = response_schema_type()
            else:
                schema = Schema.factory(type_name, {})()

        self._insert_field_numbers(schema)
        return schema

    def _insert_field_numbers(self, schema):
        counter = 1
        if ID in schema.fields:
            schema.fields[ID].meta['field_no'] = counter
            counter += 1
        if REV in schema.fields:
            schema.fields[REV].meta['field_no'] = counter
            counter += 1
        for f in schema.fields.values():
            if f.name not in {ID, REV}:
                f.meta['field_no'] = counter
                counter += 1

    def _build_request_schema(self):
        type_name = f'{StringUtils.camel(self.name)}RequestSchema'
        obj = self.decorator.kwargs.get('request')

        if isinstance(obj, dict):
            schema = Schema.factory(type_name, obj)()
        elif isinstance(obj, Schema):
            request_schema_type = type(type_name, (type(obj), ), {})
            schema = request_schema_type()
        elif is_resource_type(obj):
            request_schema_type = type(type_name, (obj.Schema, ), {})
            schema = request_schema_type()
        else:
            fields = self._infer_request_fields()
            schema = Schema.factory(type_name, fields)()

        self._insert_field_numbers(schema)
        return schema

    def _infer_request_fields(self):
        fields = {}
        for param in self.signature.parameters.values():
            field = self._infer_field(param.annotation, name=param.name)
            if field is not None:
                fields[param.name] = field
        return fields

    def _infer_field(self, annotation, name=None):
        many, type_name = extract_res_info_from_annotation(annotation)
        field = None

        if type_name in self.app.manifest.resource_classes:
            if many:
                field = field_types.List(self.app[type_name].Schema())
            else:
                field = field_types.Nested(self.app[type_name].Schema())
        elif annotation is not None:
            field_type = self._py_type_2_field_type.get(annotation)
            if field_type:
                if many:
                    field = field_types.List(field_type())
                else:
                    field = field_type()
            elif annotation is not inspect._empty:
                console.error(
                    message=f'cannot infer protobuf field from annotation',
                    data={
                        'annotation': str(annotation),
                        'name': name,
                    }
                )

        if field is not None and name:
            field.name = name

        return field



import asyncio

import uvloop

from typing import Type, Coroutine

from ravel.app.base import Application, AsyncAction


asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())


class AsyncServer(Application):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.server = None
        self.loop = None

    @property
    def action_type(self) -> Type['Action']:
        return AsyncAction

    def on_bootstrap(self, server):
        self.server = server

    def on_start(self):
        print('>>> Entering async server loop...')
        try:
            self.loop = asyncio.get_event_loop()
            self.loop.run_until_complete(self.server)
            self.loop.run_forever()
        finally:
            self.loop.close()



from ravel.util import get_class_name
from ravel.schema import Schema


def get_stripped_schema_name(obj):
    if isinstance(obj, Schema):
        name = get_class_name(obj)
    else:
        assert isinstance(obj, str)
        name = obj
    if name.endswith('Schema'):
        return name[:-len('Schema')]
    else:
        return name


from .abstract_http_server import AbstractHttpServer


class AbstractWsgiService(AbstractHttpServer):
    def on_start(self):
        def wsgi_entrypoint(environ=None, start_response=None, *args, **kwargs):
            raise NotImplementedError('override in subclass')

        return wsgi_entrypoint



import re
import random

from typing import Type

import pytest

from appyratus.utils.dict_utils import DictUtils

from ravel.test.domains.things import Thing as BaseThing
from ravel.constants import REV, ID
from ravel import Application, Store, Resource

__all__ = [
    'ResourceCrudTestSuite',
    'Thing',
    'random_things',
    'app',
]


@pytest.fixture(scope='function')
def app():
    return Application().bootstrap()


@pytest.fixture(scope='function')
def Thing(app):
    class Thing(BaseThing):
        pass

    app.register_resource(Thing)
    return Thing


@pytest.fixture(scope='function')
def random_things(Thing):
    return Thing.Batch.generate(count=64).merge(anything=1)


class ResourceCrudTestSuite:

    @classmethod
    def build_store(cls, app: Application) -> Store:
        """
        Return a bootstrapped Store object.
        """
        raise NotImplementedError('override in subclass')

    @classmethod
    def bind_store(cls, resource_type, store) -> Store:
        """
        Bind the Store with the Resource
        """
        store.bind(resource_type)

    @classmethod
    def bind(cls, resource_type: Type[Resource]):
        cls.store = cls.build_store(resource_type.ravel.app)
        cls.bind_store(resource_type, cls.store)
        resource_type.ravel.local.store = cls.store

    def test_create(self, Thing, random_things):
        self.bind(Thing)

        for thing in random_things:
            assert thing.dirty

            old_state = thing.internal.state.copy()
            thing.create()
            new_state = thing.internal.state.copy()

            # created objects should have no dirty fields
            assert not thing.dirty

            # comparing state before and after create()
            for k, v_old in old_state.items():
                v_new = new_state[k]

                if k == REV:
                    # rev is always generated by the Store
                    assert v_new is not None

                    # make sure rev string has correct structure
                    assert re.match(f'[a-z0-9]+-[a-z0-9]+', v_new)
                else:
                    # make sure field value hasn't changed following create
                    assert v_old == v_new

    def test_create_with_no_ids(self, Thing, random_things):
        self.bind(Thing)

        # make sure the store generates an ID if
        # none was passed into the create method
        for thing in random_things:
            thing._id = None
            assert thing._id is None
            thing.create()
            assert thing._id is not None

    def test_create_with_null_not_nullable_fields(self, Thing, random_things):
        Thing.name.resolver.nullable = False
        Thing.name.resolver.field.nullable = False

        random_things.name = None
        for thing in random_things:
            with pytest.raises(Exception):
                thing.create()

    def test_update(self, Thing, random_things):
        self.bind(Thing)

        for idx, thing in enumerate(random_things):
            thing.create()
            assert not thing.dirty

            if idx % 2:
                # set ever other thing's rev to make sure
                # that, following update, this erroneous value
                # is discoarded and the new real rev is set.
                thing._rev = 'foobar'

            # calling mark with no arguments makes *all* fields
            # in its state dict dirty
            thing.mark()
            assert thing.dirty

            # call update and ensure that nothing is dirty
            thing.update()
            assert not thing.dirty
            if idx % 2:
                assert thing._rev != 'foobar'

    def test_get(self, Thing, random_things):
        self.bind(Thing)

        Thing.create_many(random_things)
        all_field_names = list(Thing.Schema.fields.keys())

        for thing in random_things:
            field_names = {
                random.choice(all_field_names) for i in
                range(len(all_field_names))
            }

            thing_copy = Thing.get(thing._id, field_names)

            assert thing_copy is not None
            assert not thing_copy.dirty
            assert ID in thing_copy.internal.state
            assert REV in thing_copy.internal.state
            assert (
                set(thing_copy.internal.state.keys())
                    == (field_names | {REV, ID})
            )

    def test_get_without_fields_kwarg(self, Thing, random_things):
        self.bind(Thing)

        Thing.create_many(random_things)

        for thing in random_things:
            thing_copy = Thing.get(thing._id)
            assert thing_copy is not None
            assert not thing_copy.dirty
            assert ID in thing_copy.internal.state
            assert REV in thing_copy.internal.state

    def test_get_many(self, Thing, random_things):
        self.bind(Thing)

        Thing.create_many(random_things)

        all_field_names = list(Thing.Schema.fields.keys())
        thing_ids = [thing._id for thing in random_things]

        for _ in range(10):
            field_names = {
                random.choice(all_field_names) for i in
                range(len(all_field_names))
            }

            thing_copies = Thing.get_many(thing_ids, field_names)

            assert thing_copies
            assert not any(x.dirty for x in thing_copies)
            assert len(set(thing_copies._id)) == len(random_things)

            for thing_copy in thing_copies:
                assert thing_copy._rev is not None
                assert (
                    set(thing_copy.internal.state.keys())
                        == (field_names | {REV, ID})
                )

    def test_get_many_without_fields_kwarg(self, Thing, random_things):
        self.bind(Thing)

        Thing.create_many(random_things)

        thing_copies = Thing.get_many(random_things._id)

        assert thing_copies
        assert not any(x.dirty for x in thing_copies)
        assert all(x._id for x in thing_copies)
        assert all(x._rev for x in thing_copies)

    def test_delete(self, Thing, random_things):
        self.bind(Thing)

        Thing.create_many(random_things)

        for thing in random_things:
            thing_copy = Thing.get(thing._id, {ID})
            assert thing_copy is not None

            thing.delete()

            thing_copy = Thing.get(thing._id, {ID})

            assert thing_copy is None
            assert Thing.Schema.fields.keys() == thing.dirty.keys()

    def test_delete_many(self, Thing, random_things):
        self.bind(Thing)

        for _ in range(10):
            Thing.create_many(random_things)
            thing_ids = random_things._id

            Thing.delete_many(random_things)

            thing_copies = Thing.get_many(thing_ids, {ID})

            assert not any(x._id for x in thing_copies)
            assert not any(x._rev for x in thing_copies)
            assert all(x.dirty for x in thing_copies)

    def test_exists(self, Thing, random_things):
        self.bind(Thing)

        for thing in random_things:
            assert not Thing.exists(thing)

        Thing.create_many(random_things)

        for thing in random_things:
            assert Thing.exists(thing)

    def test_exists_many(self, Thing, random_things):
        self.bind(Thing)

        for thing in random_things:
            assert not Thing.exists(thing)
            if random.randint(0, 1):
                thing.create()

        exists = Thing.exists_many(random_things._id)

        for thing in random_things:
            if thing.dirty:
                assert exists[thing._id] is False
            else:
                assert exists[thing._id] is True

    def test_create_many(self, Thing, random_things):
        self.bind(Thing)

        Thing.create_many(random_things)
        thing_copies = Thing.get_many(random_things._id)

        assert not any(x.dirty for x in thing_copies)
        assert all(x._id for x in thing_copies)
        assert all(x._rev for x in thing_copies)

    def test_update_many(self, Thing, random_things):
        self.bind(Thing)

        Thing.create_many(random_things)

        # collect revs prior to updating all
        old_revs = set(random_things._rev)

        # make each thing dirty
        for thing in random_things:
            thing.merge(Thing.generate(), anything=1)

        # call func under test, fetch updated resources
        Thing.update_many(random_things)
        thing_copies = Thing.get_many(random_things._id)

        # everything should have an id and rev and be clean
        assert not any(x.dirty for x in thing_copies)
        assert all(x._id for x in thing_copies)
        assert all(x._rev for x in thing_copies)

        # all new revs should have been generated
        assert not (old_revs & set(thing_copies._rev))

    def test_save(self, Thing, random_things):
        self.bind(Thing)
        for thing in random_things:
            if random.randint(0, 1):
                thing.save()
            else:
                thing.create()
                thing.merge(Thing.generate(), anything=1).save()

            assert not thing.dirty
            assert thing._id is not None
            assert thing._rev is not None

    def test_save_many(self, Thing, random_things):
        self.bind(Thing)

        id_2_old_revs = {}

        # save a random subset of things
        for thing in random_things:
            if random.randint(0, 1):
                thing.create()
                id_2_old_revs[thing._id] = thing._rev

        # now call save, expecting some update calls
        # and some create calls internally.
        Thing.save_many(random_things)
        thing_copies = Thing.get_many(random_things._id)

        assert not any(x.dirty for x in thing_copies)
        assert all(x._id for x in thing_copies)
        assert all(x._rev for x in thing_copies)

        # make sure new revs are generated for
        # the merely updated resources, for both
        # in-memory copies and newly fetch ones.
        for thing in random_things:
            if thing._id in id_2_old_revs:
                assert id_2_old_revs[thing._id] != thing._rev
        for thing in thing_copies:
            if thing._id in id_2_old_revs:
                assert id_2_old_revs[thing._id] != thing._rev

    def test_query_no_params(self, Thing, random_things):
        self.bind(Thing)

        Thing.create_many(random_things)
        query = Thing.select(Thing._id)
        results = query.execute()

        assert len(results) == len(random_things)
        assert not any(x.dirty for x in results)
        assert set(results._id) == set(random_things._id)
        assert set(results._rev) == set(random_things._rev)

        for thing in results:
            assert thing.internal.state.keys() == {
                ID, REV
            }

    def test_query_with_where_clause(self, Thing, random_things):
        self.bind(Thing)

        Thing.create_many(random_things)
        query = Thing.select(Thing._id).where(Thing._id == random_things[0]._id)
        results = query.execute()

        assert len(results) == 1
        assert not any(x.dirty for x in results)

        queried_thing = random_things[0]
        result_thing = results[0]

        assert result_thing._id == queried_thing._id
        assert result_thing._rev == queried_thing._rev

    def test_query_with_limit_and_offset(self, Thing, random_things):
        self.bind(Thing)

        Thing.create_many(random_things)

        visited_ids = set()
        for offset in range(len(random_things)):
            # query each resource one by one
            query = Thing.select(Thing._id).limit(1).offset(offset)
            results = query.execute()

            assert len(results) == 1

            result = results[0]

            # make sure we haven't already seen its id
            assert result._id not in visited_ids

            visited_ids.add(result._id)

    @pytest.mark.parametrize('desc', [True, False, None])
    def test_query_with_order_by(self, Thing, random_things, desc):
        self.bind(Thing)

        Thing.create_many(random_things)

        if desc is True:
            # ensure results are ordered descending order or rev string
            query = Thing.select(Thing._id).order_by(Thing._rev.desc)
            results = query.execute()

            assert len(results) == len(random_things)
            for thing_1, thing_2 in zip(results, results[1:]):
                assert thing_1._rev >= thing_2._rev
        else:
            # ensure results are ordered ascending order or rev string
            if desc is False:
                query = Thing.select(Thing._id).order_by(Thing._rev.asc)
            else:
                query = Thing.select(Thing._id).order_by(Thing._rev)
            results = query.execute()

            assert len(results) == len(random_things)
            for thing_1, thing_2 in zip(results, results[1:]):
                assert thing_1._rev <= thing_2._rev



from typing import (
    Dict,
    Text,
)

from appyratus.usage import (
    BaseUsage,
    BaseUsageRenderer,
)
from appyratus.utils.dict_utils import DictUtils
from appyratus.utils.string_utils import StringUtils

from ravel.apps.cli import CliApplication


class ApplicationUsage(BaseUsage):

    def __init__(
        self,
        package: Text = None,
        apps: Text = None,
        entrypoint: Text = None,
        action=None,
        fields=None,
        data: Dict = None,
        **kwargs
    ):
        super().__init__(**kwargs)
        self._package = package
        self._apps = apps
        self._entrypoint = entrypoint
        self._action = action
        self._data = data
        self._fields = fields

    def render(self, context: Dict = None, **kwargs):
        base_context = {
            'fields': self._fields,
            'data': self._data,
        }
        context = self._merge_context(base_context, context)
        res = super().render(context=context, **kwargs)
        return res


class ApplicationUsageRenderer(BaseUsageRenderer):
    pass


class CliApplicationUsageRenderer(ApplicationUsageRenderer):
    """
    # Cli Application Usage Renderer
    Renders command line code for shell interpreters
    """

    SPLIT = " \\\n"
    INDENT = "  "

    def __init__(
        self, split_args=False, split_kwargs=True, indent_after_command=True, **kwargs
    ):
        """
        # Init
        # Args 
        - `split_args`, split up each argument into a new line
        - `split_kwargs`, split up each kwarg into a new line
        - `indent_after_command`, Following the command on the first line,
          indent subsequent elements into a new line with proper escape
        """
        self._split_args = split_args
        self._split_kwargs = split_kwargs
        self._indent_after_command = indent_after_command
        super().__init__(**kwargs)

    @classmethod
    def get_template(cls):
        return "{cli_display}"

    def perform(self, context: Dict = None, **kwargs):
        if context is None:
            context = {}
        ctx = self.get_context()
        ctx.update(
            {
                'app': 'cli',
                'entrypoint': StringUtils.dash(self.usage._package.__name__),
                'action': StringUtils.dash(self.usage._action.__name__),
            }
        )
        ctx.update(context)

        # build a shell command
        # the first item is the command
        # the remaining items are arguments and kwarguments
        command = ctx['entrypoint']
        args = self.build_args([ctx['app'], ctx['action']] + ctx['args'])
        kwargs = self.build_kwargs(ctx['kwargs'])
        command_line = [command]
        command_line.extend(args)
        command_line.extend(kwargs)

        # optionally indent every line but the first
        if self._indent_after_command:
            command_line[1:] = [f'{self.INDENT}{k}' for k in command_line[1:]]
        # set the value of cli display for when rendering the template
        ctx['cli_display'] = self.SPLIT.join(command_line)
        # finally render the template
        return super().perform(context=ctx)

    @classmethod
    def get_context(cls):
        return {
            'app': None,
            'entrypoint': None,
            'action': None,
            'args': [],
            'kwargs': {},
        }

    def build_args(self, args=None):
        """
        # Build Args
        """
        if not args:
            return []
        if not self._split_args:
            return [' '.join(args)]
        return args

    def build_kwargs(self, kwargs=None, split_newline=True):
        """
        # Build Kwargs
        """
        if not kwargs:
            return []
        flat_kwargs = DictUtils.flatten_keys(kwargs)
        kwargz = ['--{} {}'.format(k, v) for k, v in flat_kwargs.items() if v is not None]
        if not self._split_kwargs:
            return [' '.join(kwargz)]
        return kwargz



from typing import Callable, Tuple, Dict

from .middleware import Middleware


class ManageSession(Middleware):
    def __init__(self, getter: Callable = None, setter: Callable = None):
        self._getter = getter
        self._setter = setter

    def pre_request(
        self,
        action: 'Action',
        request: 'Request',
        raw_args: Tuple,
        raw_kwargs: Dict
    ):
        if self._getter is not None:
            request.context.session = self._getter(request)

    def post_request(
        self,
        action: 'Action',
        request: 'Request',
        result,
    ):
        if self._setter is not None:
            self._setter(request)
        if request.session is not None:
            request.session.save()
